
# Copyright (c) Microsoft Corporation.  All rights reserved.
#
# Run tests using both RxSpark/RxHadoopMR and local compute contexts
#
###############################################################################

# HadoopIOQ_SetupAndRun.R creates the getOption("rxTestArgs") required by these functions

###############################################################################
# Function to get compute contexts
###############################################################################
"checkTestArgs" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    if (is.null(rxTestArgs)) {
        stop("Compute context information must be in 'rxTestArgs' option. See HadoopIOQ_SetupAndRun.R.")
    }
}

"getIsOnNode" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    onClusterNode <- rxTestArgs$onClusterNode
    if (!is.null(onClusterNode) && isTRUE(onClusterNode)) TRUE else FALSE
}

"getHadoopComputeContext" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)
    platform <- rxTestArgs$platform

    if (!is.null(platform) && platform == "spark") {
        ## rxSparkConnect method enhances RxSpark with persistent run mode and in-memory data caching. 
        ## It is recommended to use rxSparkConnect() to run Spark application.
        ## TODO (lgong) 2017-03-22: however it changes the current CC when passed to a function via the computeContext arg!
        RxSpark( #rxSparkConnect( #
            nameNode = rxTestArgs$nameNode,
            port = rxTestArgs$port,
            hdfsShareDir = rxTestArgs$hdfsShareDir,
            shareDir = rxTestArgs$shareDir,
            sshUsername = rxTestArgs$sshUsername,
            sshHostname = rxTestArgs$sshHostname,
            sshSwitches = rxTestArgs$sshSwitches,
            wait = TRUE,
            consoleOutput = rxTestArgs$consoleOutput,
            fileSystem = rxTestArgs$fileSystem,
            sshProfileScript = rxTestArgs$sshProfileScript,
            sshClientDir = rxTestArgs$sshClientDir,
            jobTrackerURL = rxTestArgs$jobTrackerURL,
            onClusterNode = rxTestArgs$onClusterNode,
            showOutputWhileWaiting = rxTestArgs$showOutputWhileWaiting,
            autoCleanup = rxTestArgs$autoCleanup,
            workingDir = rxTestArgs$workingDir,
            dataPath = rxTestArgs$dataPath,
            outDataPath = rxTestArgs$outDataPath,
            packagesToLoad = rxTestArgs$packagesToLoad,
            numExecutors = rxTestArgs$numExecutors,
            executorCores = rxTestArgs$executorCores,
            executorMem = rxTestArgs$executorMem,
            driverMem = rxTestArgs$driverMem,
            executorOverheadMem = rxTestArgs$executorOverheadMem,
            resultsTimeout = rxTestArgs$resultsTimeout)
    } else {
        RxHadoopMR(
            nameNode = rxTestArgs$nameNode,
            port = rxTestArgs$port,
            hadoopSwitches = rxTestArgs$hadoopSwitches,
            hdfsShareDir = rxTestArgs$hdfsShareDir,
            shareDir = rxTestArgs$shareDir,
            sshUsername = rxTestArgs$sshUsername,
            sshHostname = rxTestArgs$sshHostname,
            sshSwitches = rxTestArgs$sshSwitches,
            wait = TRUE,
            consoleOutput = rxTestArgs$consoleOutput,
            fileSystem = rxTestArgs$fileSystem,
            sshProfileScript = rxTestArgs$sshProfileScript,
            hadoopRPath = rxTestArgs$hadoopRPath,
            revoPath = rxTestArgs$revoPath,
            sshClientDir = rxTestArgs$sshClientDir,
            usingRunAsUserMode = rxTestArgs$usingRunAsUserMode,
            jobTrackerURL = rxTestArgs$jobTrackerURL,
            onClusterNode = rxTestArgs$onClusterNode,
            showOutputWhileWaiting = rxTestArgs$showOutputWhileWaiting,
            autoCleanup = rxTestArgs$autoCleanup,
            workingDir = rxTestArgs$workingDir,
            dataPath = rxTestArgs$dataPath,
            outDataPath = rxTestArgs$outDataPath,
            packagesToLoad = rxTestArgs$packagesToLoad,
            resultsTimeout = rxTestArgs$resultsTimeout)
    }
}

###############################################################################
# Functions to get data sources
###############################################################################
## hdfs utilities
"getIOQHdfsPath" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)
    file.path(rxTestArgs$hdfsShareDir, rxTestArgs$myHdfsIOQSubdir)
}

"getHdfsPathWithPrefix" <- function(hdfsPath, rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    nameNode <- rxTestArgs$nameNode
    nameNodePrefix <- if (is.null(nameNode) || nameNode == "default") "" else nameNode

    paste0(nameNodePrefix, hdfsPath)
}

"removeHdfsPath" <- function(hdfsPath, skipTrash = TRUE)
{
    rxHadoopRemoveDir(getHdfsPathWithPrefix(hdfsPath), skipTrash = skipTrash)
}

"existsHdfsPath" <- function(hdfsPath)
{
    rxHadoopFileExists(getHdfsPathWithPrefix(hdfsPath))
}

"makeHdfsPath" <- function(hdfsPath)
{
    rxHadoopMakeDir(getHdfsPathWithPrefix(hdfsPath))
}

## hdfs input data
"getAirDemoCsvHdfsDS" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    airColInfo <- list(
        ArrDelay = list(type = "integer"),
        DayOfWeek = list(type = "factor", 
            levels = c("Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday")))

    RxTextData(
        file = file.path(getIOQHdfsPath(), rxTestArgs$myHdfsAirDemoCsvSubdir),
        missingValueString = "M",
        colInfo = airColInfo,
        fileSystem = rxTestArgs$fileSystem)
}

"getAirDemoParquetHdfsDS" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    airColInfo <- list(
        ArrDelay = list(type = "integer"),
        DayOfWeek = list(type = "factor",
            levels = c("Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday")))

    RxParquetData(
        file = file.path(getIOQHdfsPath(), rxTestArgs$myHdfsAirDemoParquetSubdir),
        colInfo = airColInfo,
        fileSystem = rxTestArgs$fileSystem)
}

"getAirDemoOrcHdfsDS" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    airColInfo <- list(
        ArrDelay = list(type = "integer"),
        DayOfWeek = list(type = "factor",
            levels = c("Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday")))

    RxOrcData(
        file = file.path(getIOQHdfsPath(), rxTestArgs$myHdfsAirDemoOrcSubdir),
        colInfo = airColInfo,
        fileSystem = rxTestArgs$fileSystem)
}

"getAirDemoHiveHdfsDS" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    airColInfo <- list(
        ArrDelay = list(type = "integer"),
        DayOfWeek = list(type = "factor",
            levels = c("Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday")))

    RxHiveData(
        table = rxTestArgs$myHdfsAirDemoHiveTable,
        colInfo = airColInfo)
}

"getAirDemoXdfHdfsDS" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    RxXdfData(
        file = file.path(getIOQHdfsPath(), rxTestArgs$myHdfsAirDemoXdfSubdir),
        fileSystem = rxTestArgs$fileSystem)
}

## hdfs output data
"getTestOutFileHdfsDS" <- function(testFileDirName = "", rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    RxXdfData(
        file = file.path(getIOQHdfsPath(), rxTestArgs$myHdfsTestOuputSubdir, testFileDirName),
        fileSystem = rxTestArgs$fileSystem)
}

"removeTestOutFileHdfsDS" <- function(testDS)
{
    testPath <- testDS@file
    if (existsHdfsPath(testPath)) removeHdfsPath(testPath)
}

###############################################################################
## local input data
"getAirDemoTextLocalDS" <- function()
{
    airColInfo <- list(
        ArrDelay = list(type = "integer"),
        DayOfWeek = list(type = "factor",
            levels = c("Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday")))

    RxTextData(
        file = file.path(rxGetOption("sampleDataDir"), "AirlineDemoSmall.csv"),
        colInfo = airColInfo,
        missingValueString = "M",
        fileSystem = "native")
}

"getAirDemoXdfCompositeLocalDS" <- function()
{
    RxXdfData(
        file = file.path(rxGetOption("sampleDataDir"), "AirlineDemoSmallComposite"),
        createCompositeSet = TRUE,
        fileSystem = "native")
}

"getAirDemoXdfLocalDS" <- function()
{
    RxXdfData(
        file = file.path(rxGetOption("sampleDataDir"), "AirlineDemoSmall.xdf"),
        fileSystem = "native")
}

## local output data
"getTestOutFileLocalDS" <- function(testFileDirName = "", rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    RxXdfData(
        file = file.path(rxTestArgs$localTestDataDir, testFileDirName),
        createCompositeSet = TRUE,
        fileSystem = "native")
}

"removeTestOutFileLocalDS" <- function(testDS)
{
    testPath <- testDS@file
    if (file.exists(testPath)) unlink(testPath, recursive = TRUE)
}

#############################################################################################
# Tests to put data on Hadoop cluster
#############################################################################################

"test.hadoop.aaa.initializeHadoop" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    if (.Platform$OS != "windows") {
        shortPathName <- function(path) path
    }

    oldCC <- rxSetComputeContext(getHadoopComputeContext())
    on.exit(rxSetComputeContext(oldCC), add = TRUE)

    # Get the curent version from the Hadoop cluster
    versionOutput <- rxExec(function() Revo.version)
    cat("MRS version on Hadoop cluster: ", as.character(versionOutput$rxElem1$version.string), '\n')
    cat("Hadoop Compute Context being used: ",
        capture.output(print(getHadoopComputeContext())),
        '\n', sep = '\n')

    if (!existsHdfsPath(rxTestArgs$hdfsShareDir)) {
        cat("WARNING: hdfsShareDir does not exist or problem with compute context.\n")
    }

    if (!rxTestArgs$createHadoopData) return(invisible(T))

    # ------------------------------------------------------------------
    # csv
    # Destination path for files in Hadoop
    airDemoSmallCsvHdfsPath <- getAirDemoCsvHdfsDS()@file

    # --- Remove the directory if it's already there ---
    if (existsHdfsPath(airDemoSmallCsvHdfsPath)) {
        cat("Removing `", airDemoSmallCsvHdfsPath, "'\n", sep = "")
        removeHdfsPath(airDemoSmallCsvHdfsPath)
    }

    # ... and now re-create it so that we can copy files into it
    makeHdfsPath(airDemoSmallCsvHdfsPath)

    # --- copy split CSV data set ---
    airDemoSmallSplitDir <- file.path(rxGetOption("sampleDataDir"), "AirlineDemoSmallSplit")
    files <- list.files(airDemoSmallSplitDir, recursive = TRUE, include.dirs = FALSE, full.names = TRUE)
    checkEquals(length(files), 3)

    if (getIsOnNode()) {
        rc <- lapply(files, function(src) {
            rxHadoopCopyFromLocal(source = src, dest = getHdfsPathWithPrefix(airDemoSmallCsvHdfsPath))
        })
    } else {
        # Copy the file from local client
        rc <- lapply(files, function(src) {
            rxHadoopCopyFromClient(source = file.path(shortPathName(dirname(src)), basename(src)),
                nativeTarget = rxTestArgs$myHdfsNativeDir,
                hdfsDest = getHdfsPathWithPrefix(airDemoSmallCsvHdfsPath),
                computeContext = getHadoopComputeContext(),
                sshUsername = rxTestArgs$sshUsername,
                sshHostname = rxTestArgs$sshHostname,
                sshSwitches = rxTestArgs$sshSwitches)
        })
    }

    checkTrue(all(unlist(rc)))

    # ------------------------------------------------------------------
    # parquet
    # Destination path for files in Hadoop
    airDemoSmallParquetHdfsPath <- getAirDemoParquetHdfsDS()@file

    # --- Remove the directory if it's already there ---
    if (existsHdfsPath(airDemoSmallParquetHdfsPath)) {
        cat("Removing `", airDemoSmallParquetHdfsPath, "'\n", sep = "")
        removeHdfsPath(airDemoSmallParquetHdfsPath)
    }

    # ... and now re-create it so that we can copy files into it
    makeHdfsPath(airDemoSmallParquetHdfsPath)

    # --- copy Parquet data set ---
    airDemoSmallParquetDir <- file.path(rxGetOption("sampleDataDir"), "AirlineDemoSmallParquet")
    files <- list.files(airDemoSmallParquetDir, recursive = TRUE, include.dirs = FALSE, full.names = TRUE)
    checkEquals(length(files), 6)

    if (getIsOnNode()) {
        rc <- lapply(files, function(src) {
            rxHadoopCopyFromLocal(source = src, dest = getHdfsPathWithPrefix(airDemoSmallParquetHdfsPath))
        })
    } else {
        # Copy the file from local client
        rc <- lapply(files, function(src) {
            rxHadoopCopyFromClient(source = file.path(shortPathName(dirname(src)), basename(src)),
                nativeTarget = rxTestArgs$myHdfsNativeDir,
                hdfsDest = getHdfsPathWithPrefix(airDemoSmallParquetHdfsPath),
                computeContext = getHadoopComputeContext(),
                sshUsername = rxTestArgs$sshUsername,
                sshHostname = rxTestArgs$sshHostname,
                sshSwitches = rxTestArgs$sshSwitches)
        })
    }

    checkTrue(all(unlist(rc)))

    # ------------------------------------------------------------------
    # orc
    # Destination path for files in Hadoop
    airDemoSmallOrcHdfsPath <- getAirDemoOrcHdfsDS()@file

    # --- Remove the directory if it's already there ---
    if (existsHdfsPath(airDemoSmallOrcHdfsPath)) {
        cat("Removing `", airDemoSmallOrcHdfsPath, "'\n", sep = "")
        removeHdfsPath(airDemoSmallOrcHdfsPath)
    }

    # ... and now re-create it so that we can copy files into it
    makeHdfsPath(airDemoSmallOrcHdfsPath)

    # --- copy Orc data set ---
    airDemoSmallOrcDir <- file.path(rxGetOption("sampleDataDir"), "AirlineDemoSmallOrc")
    files <- list.files(airDemoSmallOrcDir, recursive = TRUE, include.dirs = FALSE, full.names = TRUE)
    checkEquals(length(files), 4)

    if (getIsOnNode()) {
        rc <- lapply(files, function(src) {
            rxHadoopCopyFromLocal(source = src, dest = getHdfsPathWithPrefix(airDemoSmallOrcHdfsPath))
        })
    } else {
        # Copy the file from local client
        rc <- lapply(files, function(src) {
            rxHadoopCopyFromClient(source = file.path(shortPathName(dirname(src)), basename(src)),
                nativeTarget = rxTestArgs$myHdfsNativeDir,
                hdfsDest = getHdfsPathWithPrefix(airDemoSmallOrcHdfsPath),
                computeContext = getHadoopComputeContext(),
                sshUsername = rxTestArgs$sshUsername,
                sshHostname = rxTestArgs$sshHostname,
                sshSwitches = rxTestArgs$sshSwitches)
        })
    }

    checkTrue(all(unlist(rc)))

    # ------------------------------------------------------------------
    # hive
    # Destination path for files in Hadoop
    dataHiveHdfs <- try(rxDataStep(inData = getAirDemoCsvHdfsDS(), outFile = getAirDemoHiveHdfsDS(), overwrite = TRUE))
    checkTrue(!inherits(dataHiveHdfs, "try-error"))

    # ------------------------------------------------------------------
    # cxdf
    # copy composite XDF data set
    airDemoSmallXdfHdfsPath <- getAirDemoXdfHdfsDS()@file
    localPath <- file.path(rxGetOption("sampleDataDir"), "AirlineDemoSmallComposite")

    if (existsHdfsPath(airDemoSmallXdfHdfsPath)) {
        cat('removing `', airDemoSmallXdfHdfsPath, "'\n", sep="")
        removeHdfsPath(airDemoSmallXdfHdfsPath)
    }

    # (1) first create the directory structure in the destination
    dirs <- unique(dirname(list.files(localPath, recursive = TRUE, include.dirs = FALSE)))
    rc <- lapply(dirs, function (subdir) {
        makeHdfsPath(file.path(airDemoSmallXdfHdfsPath, subdir))
    })

    checkTrue(all(unlist(rc)))

    # (2) then copy 4 data/metadata files
    files <- list.files(localPath, recursive = TRUE, include.dirs = FALSE)
    checkEquals(length(files), 4) # 3 data files + 1 metadata

    if (getIsOnNode()) {
        rc <- lapply(files, function (subPath) {
            src <- file.path(localPath, subPath)
            dst <- file.path(airDemoSmallXdfHdfsPath, subPath)
            rxHadoopCopyFromLocal(src, getHdfsPathWithPrefix(dst))
        })
    } else {
        # Copy the file from local client
        rc <- lapply(files, function(subPath) {
            src <- file.path(localPath, subPath)
            dst <- file.path(airDemoSmallXdfHdfsPath, subPath)
            rxHadoopCopyFromClient(source = file.path(shortPathName(dirname(src)), basename(src)), 
                nativeTarget = rxTestArgs$myHdfsNativeDir,
                hdfsDest = getHdfsPathWithPrefix(dst),
                computeContext = getHadoopComputeContext(),
                sshUsername = rxTestArgs$sshUsername,
                sshHostname = rxTestArgs$sshHostname,
                sshSwitches = rxTestArgs$sshSwitches)
        })
    }

    checkTrue(all(unlist(rc)))

    # --- Make directory in HDFS for writing temporary test data -------
    outDS <- getTestOutFileHdfsDS()
    rc <- makeHdfsPath(outDS@file)

    checkTrue(rc)
}

"getLocalComputeContext" <- function()
{
    RxLocalSeq()
}

#############################################################################################
# Tests for analysis functions that read data
#############################################################################################
"test.hadoop.rxEnsemble.fastTrees.regression.localseq.csv" <- function()
{
    oldCC <- rxSetComputeContext(RxLocalSeq())
    dataCsvLocal <- getAirDemoTextLocalDS()

    ###################################
    ## RxLocalSeq, splitData = FALSE
    ensemble.local.seq <- try(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataCsvLocal
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111
        , replace = FALSE
        , modelCount = 4
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = FALSE
        ))
    checkTrue(!inherits(ensemble.local.seq, "try-error"))
    checkTrue(inherits(rxGetComputeContext(), "RxLocalSeq"))

    ## rxPredict
    outLocalSeqDS <- getTestOutFileLocalDS("outLocalSeqDS")
    on.exit(removeTestOutFileLocalDS(outLocalSeqDS))
    scoresLocalSeq <- rxPredict(ensemble.local.seq, data = dataCsvLocal, outData = outLocalSeqDS
        , overwrite = TRUE, writeModelVars = TRUE, extraVarsToWrite = "CRSDepTime", suffix = "Pred"
        )
    scoresLocalSeqInfo <- rxGetInfo(scoresLocalSeq, getVarInfo = TRUE, getBlockSizes = TRUE)
    checkEquals(names(scoresLocalSeqInfo$varInfo), c("CRSDepTime", "ArrDelay", "DayOfWeek", "ScorePred"))

    ###################################
    ## RxLocalSeq, splitData = TRUE 
    checkException(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataCsvLocal
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111
        , replace = FALSE
        , modelCount = 4
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = TRUE
        ))

    on.exit(rxSetComputeContext(oldCC), add = TRUE)
}

"test.hadoop.rxEnsemble.fastTrees.regression.localpar.xdf" <- function()
{
    oldCC <- rxSetComputeContext(RxLocalParallel())
    dataXdfLocal <- getAirDemoXdfLocalDS()

    ###################################
    ## RxLocalParallel, splitData = FALSE
    ensemble.local.par <- try(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataXdfLocal
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111        
        , replace = FALSE
        , modelCount = 4    # 4 != length(trainers)
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = FALSE
        ))
    checkTrue(!inherits(ensemble.local.par, "try-error"))
    checkTrue(inherits(rxGetComputeContext(), "RxLocalParallel"))

    ## rxPredict
    outLocalParDS <- getTestOutFileLocalDS("outLocalParDS")
    on.exit(removeTestOutFileLocalDS(outLocalParDS))
    scoresLocalPar <- rxPredict(ensemble.local.par, data = dataXdfLocal, outData = outLocalParDS
        , overwrite = TRUE, writeModelVars = TRUE, extraVarsToWrite = "CRSDepTime", suffix = "Pred"
        )
    scoresLocalParInfo <- rxGetInfo(scoresLocalPar, getVarInfo = TRUE, getBlockSizes = TRUE)
    checkEquals(names(scoresLocalParInfo$varInfo), c("CRSDepTime", "ArrDelay", "DayOfWeek", "ScorePred"))

    ###################################
    ## RxLocalParallel, splitData = TRUE
    checkException(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataCsvLocal
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111        
        , replace = FALSE
        , modelCount = 4    # 4 != length(trainers)
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = TRUE 
        ))

    on.exit(rxSetComputeContext(oldCC), add = TRUE)
}

"test.hadoop.rxEnsemble.fastTrees.regression.xdf" <- function()
{
    oldCC <- rxSetComputeContext(getHadoopComputeContext())
    dataXdfHdfs <- getAirDemoXdfHdfsDS()

    ###################################
    ## RxSpark with cxdf, splitData = FALSE
    ensemble.spark.xdf.F <- try(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataXdfHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111
        , replace = FALSE
        , modelCount = 4
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = FALSE
        ))
    checkTrue(!inherits(ensemble.spark.xdf.F, "try-error"))
    checkTrue(inherits(rxGetComputeContext(), class(getHadoopComputeContext())))

    ## rxPredict
    outHadoopXdfDSF <- getTestOutFileHdfsDS("outHadoopXdfDSF")
    on.exit(removeTestOutFileHdfsDS(outHadoopXdfDSF), add = TRUE)
    scoresSparkXdfF <- rxPredict(ensemble.spark.xdf.F, data = dataXdfHdfs, outData = outHadoopXdfDSF, overwrite = TRUE
        , writeModelVars = TRUE, extraVarsToWrite = "CRSDepTime", suffix = "Pred"
        )
    scoresSparkXdfFInfo <- rxGetInfo(scoresSparkXdfF, getVarInfo = TRUE, getBlockSizes = TRUE)
    checkTrue(setequal(names(scoresSparkXdfFInfo$varInfo), c("CRSDepTime", "ArrDelay", "DayOfWeek", "ScorePred")))

    ###################################
    ## RxSpark with cxdf, splitData = TRUE 
    ensemble.spark.xdf.T <- try(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataXdfHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111        
        , replace = FALSE
        , modelCount = 4    # 4 != length(trainers)
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = TRUE 
        ))
    checkTrue(!inherits(ensemble.spark.xdf.T, "try-error"))
    checkTrue(inherits(rxGetComputeContext(), class(getHadoopComputeContext())))

    ## rxPredict
    outHadoopXdfDST <- getTestOutFileHdfsDS("outHadoopXdfDST")
    on.exit(removeTestOutFileHdfsDS(outHadoopXdfDST), add = TRUE)
    scoresSparkXdfT <- rxPredict(ensemble.spark.xdf.T, data = dataXdfHdfs, outData = outHadoopXdfDST, overwrite = TRUE
        , writeModelVars = TRUE, extraVarsToWrite = "CRSDepTime", suffix = "Pred"
        )
    scoresSparkXdfTInfo <- rxGetInfo(scoresSparkXdfT, getVarInfo = TRUE, getBlockSizes = TRUE)
    checkTrue(setequal(names(scoresSparkXdfTInfo$varInfo), c("CRSDepTime", "ArrDelay", "DayOfWeek", "ScorePred")))

    on.exit(rxSetComputeContext(oldCC), add = TRUE)
}

"test.hadoop.rxEnsemble.fastTrees.regression.csv" <- function()
{
    oldCC <- rxSetComputeContext(getHadoopComputeContext())
    dataCsvHdfs <- getAirDemoCsvHdfsDS()

    ###################################
    ## RxSpark with text, splitData = FALSE
    ensemble.spark.csv.F <- try(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataCsvHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111        
        , replace = FALSE
        , modelCount = 4    # 4 != length(trainers)
        , sampRate = 0.75
        , combineMethod = NULL 
        , splitData = FALSE 
        ))
    checkTrue(!inherits(ensemble.spark.csv.F, "try-error"))
    checkTrue(inherits(rxGetComputeContext(), class(getHadoopComputeContext())))

    ## rxPredict
    outHadoopCsvDSF <- getTestOutFileHdfsDS("outHadoopCsvDSF")
    on.exit(removeTestOutFileHdfsDS(outHadoopCsvDSF), add = TRUE)
    assign("predictMethod", "useDataStep", envir = MicrosoftML:::rxHashEnv)
    scoresSparkCsvF <- rxPredict(ensemble.spark.csv.F, data = dataCsvHdfs, outData = outHadoopCsvDSF, overwrite = TRUE
        , writeModelVars = TRUE, extraVarsToWrite = "CRSDepTime", suffix = "Pred"
        )
    assign("predictMethod", "useDefault", envir = MicrosoftML:::rxHashEnv)
    scoresSparkCsvFInfo <- rxGetInfo(scoresSparkCsvF, getVarInfo = TRUE, getBlockSizes = TRUE)
    checkTrue(setequal(names(scoresSparkCsvFInfo$varInfo), c("CRSDepTime", "ArrDelay", "DayOfWeek", "ScorePred")))

    ###################################
    ## RxSpark with text, splitData = TRUE
    rxSetComputeContext(getHadoopComputeContext(), persistentRun = TRUE)
    ensemble.spark.csv.T <- try(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataCsvHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111        
        , replace = FALSE
        , modelCount = 4    # 4 != length(trainers)
        , sampRate = 0.75
        , combineMethod = NULL 
        , splitData = TRUE 
        ))
    #Error in rxExecJob(matchCallList, matchCall = matchCall) :
      #Hive temp table is an in-memory object. You can only use it in RxSpark with persistentRun mode
    checkTrue(!inherits(ensemble.spark.csv.T, "try-error"))
    checkTrue(inherits(rxGetComputeContext(), class(getHadoopComputeContext())))

    ## rxPredict
    outHadoopCsvDST <- getTestOutFileHdfsDS("outHadoopCsvDST")
    on.exit(removeTestOutFileHdfsDS(outHadoopCsvDST), add = TRUE)
    assign("predictMethod", "useDataStep", envir = MicrosoftML:::rxHashEnv)
    scoresSparkCsvT <- rxPredict(ensemble.spark.csv.T, data = dataCsvHdfs, outData = outHadoopCsvDST, overwrite = TRUE
        , writeModelVars = TRUE, extraVarsToWrite = "CRSDepTime", suffix = "Pred"
        )
    assign("predictMethod", "useDefault", envir = MicrosoftML:::rxHashEnv)
    scoresSparkCsvTInfo <- rxGetInfo(scoresSparkCsvT, getVarInfo = TRUE, getBlockSizes = TRUE)
    checkTrue(setequal(names(scoresSparkCsvTInfo$varInfo), c("CRSDepTime", "ArrDelay", "DayOfWeek", "ScorePred")))

    rxStopEngine(rxGetComputeContext())
    on.exit(rxSetComputeContext(oldCC), add = TRUE)
}

## RxSpark with parquet (RxSparkData is not supported in HPC (rxExec) mode)
#   |__ RxSparkData
#       |__ RxParquetData
#       |__ RxHiveData
#       |__ RxOrcData
"test.hadoop.rxEnsemble.fastTrees.regression.parquet" <- function()
{
    oldCC <- rxSetComputeContext(getHadoopComputeContext())
    dataParquetHdfs <- getAirDemoParquetHdfsDS()

    ###################################
    ## RxSpark with parquet, splitData = TRUE
    ensemble.spark.parquet <- try(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataParquetHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111
        , replace = FALSE
        , modelCount = 4
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = TRUE  #must be TRUE
        ))
    checkTrue(!inherits(ensemble.spark.parquet, "try-error"))
    checkTrue(inherits(rxGetComputeContext(), class(getHadoopComputeContext()))) #TODO: failed due to a bug in rxEnsemble

    ## rxPredict
    outHadoopParquetDS <- getTestOutFileHdfsDS("outHadoopParquetDS")
    on.exit(removeTestOutFileHdfsDS(outHadoopParquetDS), add = TRUE)
    scoresSparkParquet <- rxPredict(ensemble.spark.parquet, data = dataParquetHdfs, outData = outHadoopParquetDS, overwrite = TRUE
        , writeModelVars = TRUE, extraVarsToWrite = "CRSDepTime", suffix = "Pred"
        )
    scoresSparkParquetInfo <- rxGetInfo(scoresSparkParquet, getVarInfo = TRUE, getBlockSizes = TRUE)
    checkTrue(setequal(names(scoresSparkParquetInfo$varInfo), c("CRSDepTime", "ArrDelay", "DayOfWeek", "ScorePred")))

    ###################################
    ## RxSpark with parquet, splitData = FALSE
    checkException(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataParquetHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111
        , replace = FALSE
        , modelCount = 4
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = FALSE
        ))

    on.exit(rxSetComputeContext(oldCC), add = TRUE)
}

"test.hadoop.rxEnsemble.fastTrees.regression.orc" <- function()
{
    oldCC <- rxSetComputeContext(getHadoopComputeContext())
    dataOrcHdfs <- getAirDemoOrcHdfsDS()

    ##################################################
    ## RxSpark with orc, splitData = TRUE
    ensemble.spark.orc <- try(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataOrcHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111
        , replace = FALSE
        , modelCount = 4
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = TRUE
        ))
    checkTrue(!inherits(ensemble.spark.orc, "try-error"))
    checkTrue(inherits(rxGetComputeContext(), class(getHadoopComputeContext())))

    ## rxPredict
    outHadoopOrcDS <- getTestOutFileHdfsDS("outHadoopOrcDS")
    on.exit(removeTestOutFileHdfsDS(outHadoopOrcDS), add = TRUE)
    scoresSparkOrc <- rxPredict(ensemble.spark.orc, data = dataOrcHdfs, outData = outHadoopOrcDS, overwrite = TRUE
        , writeModelVars = TRUE, extraVarsToWrite = "CRSDepTime", suffix = "Pred"
        )
    scoresSparkOrcInfo <- rxGetInfo(scoresSparkOrc, getVarInfo = TRUE, getBlockSizes = TRUE)
    checkTrue(setequal(names(scoresSparkOrcInfo$varInfo), c("CRSDepTime", "ArrDelay", "DayOfWeek", "ScorePred")))

    ###################################
    ## RxSpark with orc, splitData = FALSE
    checkException(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataOrcHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111
        , replace = FALSE
        , modelCount = 4
        , sampRate = 0.75
        , combineMethod = NULL
        , splitData = FALSE
        ))

    on.exit(rxSetComputeContext(oldCC), add = TRUE)
}

"test.hadoop.rxEnsemble.fastTrees.regression.hive" <- function()
{
    oldCC <- rxSetComputeContext(getHadoopComputeContext())
    dataHiveHdfs <- getAirDemoHiveHdfsDS()

    ##################################################
    ## RxSpark with hive, splitData = TRUE
    ensemble.spark.hive <- try(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataHiveHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111        
        , replace = FALSE
        , modelCount = 4
        , sampRate = 0.75
        , combineMethod = NULL 
        , splitData = TRUE
        ))
    checkTrue(!inherits(ensemble.spark.hive, "try-error"))
    checkTrue(inherits(rxGetComputeContext(), class(getHadoopComputeContext())))

    ## rxPredict
    outHadoopHiveDS <- getTestOutFileHdfsDS("outHadoopHiveDS")
    on.exit(removeTestOutFileHdfsDS(outHadoopHiveDS), add = TRUE)
    scoresSparkHive <- rxPredict(ensemble.spark.hive, data = dataHiveHdfs, outData = outHadoopHiveDS, overwrite = TRUE
        , writeModelVars = TRUE, extraVarsToWrite = "CRSDepTime", suffix = "Pred"
        )
    scoresSparkHiveInfo <- rxGetInfo(scoresSparkHive, getVarInfo = TRUE, getBlockSizes = TRUE)
    checkTrue(setequal(names(scoresSparkHiveInfo$varInfo), c("CRSDepTime", "ArrDelay", "DayOfWeek", "ScorePred")))

    ###################################
    ## RxSpark with hive, splitData = FALSE
    checkException(rxEnsemble(
        formula = ArrDelay ~ DayOfWeek
        , data = dataHiveHdfs
        , type = "regression"
        , trainers = list(fastTrees(), fastTrees(numTrees = 60), fastTrees(learningRate = 0.1))
        , randomSeed = 111        
        , replace = FALSE
        , modelCount = 4
        , sampRate = 0.75
        , combineMethod = NULL 
        , splitData = FALSE
        ))

    on.exit(rxSetComputeContext(oldCC), add = TRUE)
}

#############################################################################################
# Always run last: Test to remove data from Hadoop cluster
#############################################################################################

"test.hadoop.zzz.removeHadoopData" <- function(rxTestArgs = getOption("rxTestArgs"))
{
    checkTestArgs(rxTestArgs)

    oldCC <- rxSetComputeContext(getHadoopComputeContext())
    on.exit(rxSetComputeContext(oldCC), add = TRUE)

    # Always remove temp data
    res <- TRUE
    if (isTRUE(rxTestArgs$removeHadoopDataOnCompletion))
    {
        ##NOTE (lgong) 2017-03-22: a hack to drop hive table, the error can be ignored.
        #dataHiveHdfs <- getAirDemoHiveHdfsDS()
        #dataHiveHdfsDropOut <- capture.output(dataHiveHdfsDrop <- try(
            #rxDataStep(RxHiveData(query = paste("DROP TABLE", dataHiveHdfs@table)), silent = TRUE)))

        # Remove IOQ (air data and temp data) from Hadoop cluster
        res <- removeHdfsPath(getIOQHdfsPath())
    }
    #else
    #{
        #res <- removeHdfsPath(getTestOutFileHdfsDS()@file)
    #}

    checkTrue(res)
}
