# Data Set: Communities and Crime
#
# Abstract: Communities within the United States. The data combines socio-economic data 
# from the 1990 US Census, law enforcement data from the 1990 US LEMAS survey, and crime 
# data from the 1995 FBI UCR.

# If 'download' is set to TRUE and the data is not available locally, 
# it will be downloaded from:
# from https://archive.ics.uci.edu/ml/machine-learning-databases/communities/
# and stored in the sampleDataDir

"getCommunitiesData" <- function(download = TRUE, sampleDataDir = NULL, createDir = TRUE)
{
    sampleDataDir <- getSampleDataDir(sampleDataDir = sampleDataDir,
        createDir = createDir) 
    
    # Check to see if xdf files exist
    commPath <- file.path(sampleDataDir, "communities")
    if (!file.exists(commPath))
    {
        dir.create(commPath)
    }
    
    trainPath <- file.path(commPath, "communitiesTrain.xdf")
    testPath <- file.path(commPath, "communitiesTest.xdf")
    if (file.exists(trainPath) && file.exists(testPath))
    {
        communities <- list(
            communitiesTrain = rxDataStep(trainPath), 
            communitiesTest = rxDataStep(testPath))
        return( communities )
    }
    if (!download)
    {
        stop(paste("Data files are not available in: ", commPath))
    }
    else
    {
        cat("Downloading and importing data to:", commPath, "\n")
    }

	label <- "ViolentCrimesPerPop"
	features <- c("state","county","community","communityname","fold", "population", "householdsize", "racepctblack", "racePctWhite", 
        "racePctAsian", "racePctHisp", "agePct12t21", "agePct12t29", "agePct16t24", "agePct65up", "numbUrban", "pctUrban", "medIncome", "pctWWage", 
        "pctWFarmSelf", "pctWInvInc", "pctWSocSec", "pctWPubAsst", "pctWRetire", "medFamInc", "perCapInc", "whitePerCap", "blackPerCap", "indianPerCap", 
        "AsianPerCap", "OtherPerCap", "HispPerCap", "NumUnderPov", "PctPopUnderPov", "PctLess9thGrade", "PctNotHSGrad", "PctBSorMore", "PctUnemployed", 
        "PctEmploy", "PctEmplManu", "PctEmplProfServ", "PctOccupManu", "PctOccupMgmtProf", "MalePctDivorce", "MalePctNevMarr", "FemalePctDiv", "TotalPctDiv", 
        "PersPerFam", "PctFam2Par", "PctKids2Par", "PctYoungKids2Par", "PctTeen2Par", "PctWorkMomYoungKids", "PctWorkMom", "NumIlleg", "PctIlleg", "NumImmig", 
        "PctImmigRecent", "PctImmigRec5", "PctImmigRec8", "PctImmigRec10", "PctRecentImmig", "PctRecImmig5", "PctRecImmig8", "PctRecImmig10", "PctSpeakEnglOnly", 
        "PctNotSpeakEnglWell", "PctLargHouseFam", "PctLargHouseOccup", "PersPerOccupHous", "PersPerOwnOccHous", "PersPerRentOccHous", "PctPersOwnOccup", "PctPersDenseHous", 
        "PctHousLess3BR", "MedNumBR", "HousVacant", "PctHousOccup", "PctHousOwnOcc", "PctVacantBoarded", "PctVacMore6Mos", "MedYrHousBuilt", "PctHousNoPhone", 
        "PctWOFullPlumb", "OwnOccLowQuart", "OwnOccMedVal", "OwnOccHiQuart", "RentLowQ", "RentMedian", "RentHighQ", "MedRent", "MedRentPctHousInc", 
        "MedOwnCostPctInc", "MedOwnCostPctIncNoMtg", "NumInShelters", "NumStreet", "PctForeignBorn", "PctBornSameState", "PctSameHouse85", "PctSameCity85", "PctSameState85", 
        "LemasSwornFT", "LemasSwFTPerPop", "LemasSwFTFieldOps", "LemasSwFTFieldPerPop", "LemasTotalReq", "LemasTotReqPerPop", "PolicReqPerOffic", "PolicPerPop", "RacialMatchCommPol", 
        "PctPolicWhite", "PctPolicBlack", "PctPolicHisp", "PctPolicAsian", "PctPolicMinor", "OfficAssgnDrugUnits", "NumKindsDrugsSeiz", "PolicAveOTWorked", "LandArea", "PopDens", 
        "PctUsePubTrans", "PolicCars", "PolicOperBudg", "LemasPctPolicOnPatr", "LemasGangUnitDeploy", "LemasPctOfficDrugUn", "PolicBudgPerPop")
	col.names <-c(features,label)
	
    communities.data.url <- "http://archive.ics.uci.edu/ml/machine-learning-databases/communities/communities.data"
    
    communities.data <- try(read.csv(url(communities.data.url), col.names = col.names, header = FALSE, na.string="?",
        check.names = FALSE, strip.white = TRUE, comment.char = "|", stringsAsFactors = FALSE))
        
    stopifnot(!inherits(communities.data, "try-error"))

    # Replace missing values with 0's and create indicator columns
    col.missingValues <- colnames(communities.data)[colSums(is.na(communities.data)) > 0]

	for(columnWithMissing in col.missingValues)
	{
		naColumn <- paste0(columnWithMissing, ".IsNa")     
        communities.data[naColumn] <- ifelse(is.na(communities.data[,columnWithMissing]), TRUE, FALSE)
        communities.data[, columnWithMissing] <- ifelse(communities.data[, naColumn], 0, communities.data[, columnWithMissing])

	}

    set.seed(12)
    splitCommunitiesData <- rxSplit(inData = communities.data, 
       splitByFactor = "splitVar", 
       transforms=list(splitVar = factor( 
        sample(0:1,  size=.rxNumRows, replace=TRUE, prob=c(.8, .2)), 
        levels=0:1, labels = c("Train", "Test"))), 
      overwrite = TRUE)
	
    numCols <- ncol(splitCommunitiesData[[1]])
    
    # Remove the temporary 'splitVar', the last column before saving to local xdf files and returning
    rxDataStep(inData = splitCommunitiesData[[1]][, -numCols], outFile = trainPath)
    rxDataStep(inData = splitCommunitiesData[[2]][, -numCols], outFile = testPath)
    
    communities <- list(communitiesTrain =  splitCommunitiesData[[1]][, -numCols], communitiesTest = splitCommunitiesData[[2]][, -numCols])
    return( communities )
}
