# Data Set: The MNIST database of handwritten digits
#
# Abstract: Communities within the United States. The data combines socio-economic data 
# from the 1990 US Census, law enforcement data from the 1990 US LEMAS survey, and crime 
# data from the 1995 FBI UCR.

# If 'download' is set to TRUE and the data is not available locally, it will be downloaded from:
# http://yann.lecun.com/exdb/mnist/
# and stored in the sampleDataDir

"getMnistData" <- function(download = TRUE, sampleDataDir = NULL, createDir = TRUE)
{
    sampleDataDir <- getSampleDataDir(sampleDataDir = sampleDataDir,
        createDir = createDir) 
    
    # Check to see if xdf files exist
    mnistPath <- file.path(sampleDataDir, "mnist")
    if (!file.exists(mnistPath))
    {
        dir.create(mnistPath)
    }
    
    trainPath <- file.path(mnistPath, "mnistTrain.xdf")
    testPath <- file.path(mnistPath, "mnistTest.xdf")
    if (file.exists(trainPath) && file.exists(testPath))
    {
        mnist <- list(
            mnistTrain = rxDataStep(trainPath, maxRowsByCols = 5e+07), 
            mnistTest = rxDataStep(testPath, maxRowsByCols = 5e+07))
        return( mnist )
    }
    if (!download)
    {
        stop(paste("Data files are not available in: ", mnistPath))
    }
    else
    {
        cat("Downloading and importing data to:", mnistPath, "\n")
    }

    `readImages` <- function(description = "")
    {
        con <- gzcon(url(description))
        magic_nimgs_nrows_ncols <- readBin(con, integer(), n = 4, endian = "big")
        nimgs <- magic_nimgs_nrows_ncols[2]
        nrows <- magic_nimgs_nrows_ncols[3]
        ncols <- magic_nimgs_nrows_ncols[4]
        
        nhead <- 4 * 4 #4 4-byte integers
        n <- nhead + nimgs * nrows * ncols
        images <- readBin(con, integer(), n = n, size = 1, signed = FALSE, endian = "big")
        images <- images[- seq.int(nhead)]
        images <- matrix(images, nrow = nimgs, byrow = TRUE)
    }

    `readLabels` <- function(description = "")
    {
        con <- gzcon(url(description))
        magic_nlabs <- readBin(con, integer(), n = 4, endian = "big")
        nlabs <- magic_nlabs[2]
        
        nhead <- 2 * 4 #2 4-byte integers
        n <- nhead + nlabs
        labels <- readBin(con, integer(), n = n, size = 1, signed = FALSE, endian = "big")
        labels <- labels[- seq.int(nhead)]
    }

    train_images_idx3_ubyte <- "http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz"
    train_images <- readImages(train_images_idx3_ubyte)
    train_labels_idx1_ubyte <- "http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz"
    train_labels <- readLabels(train_labels_idx1_ubyte)
    stopifnot(nrow(train_images) == length(train_labels))
        
    t10k_images_idx3_ubyte <- "http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz"
    t10k_images <- readImages(t10k_images_idx3_ubyte)
    t10k_labels_idx1_ubyte <- "http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz"
    t10k_labels <- readLabels(t10k_labels_idx1_ubyte)    
    stopifnot(nrow(t10k_images) == length(t10k_labels))
    
    ylab <- "Label"; xlab <- "Feature"
    image.size <- ncol(train_images)
    df.train <- setNames(data.frame(train_labels, train_images), c(ylab, paste0(xlab, seq.int(image.size))))

    # Creates the testing data frame.
    df.test <- setNames(data.frame(t10k_labels, t10k_images), c(ylab, paste0(xlab, seq.int(image.size))))        
    
    # Save to local xdf files
    rxDataStep(inData = df.train, outFile = trainPath)
    rxDataStep(inData = df.test, outFile = testPath)

    list(mnistTrain = df.train, mnistTest = df.test)
}
