# Copyright (c) Microsoft Corporation.  All rights reserved.

# Tests for basic functionality of rxFastForest

"test.rxFastForest.basic" <- function()
{
    # Create temporary data sources
    myIris = iris
    myIris$Setosa <- iris$Species == "setosa"
    myXdf <- tempfile(pattern = "myIris", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)

    rxDataStep(myIris, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)
    
    myCsv <- tempfile(pattern = "myIris", fileext = ".csv")
    on.exit(file.remove(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, colInfo = rxCreateColInfo(iris))
    rxDataStep(myIris, myCsvDS, overwrite = TRUE)
    
    readVars <- c("Setosa", "Sepal.Length", "Petal.Length", "Sepal.Width")
    
    # Run examples
    myFormula <- Setosa ~ Sepal.Length + Petal.Length + Sepal.Width

    # xdf file
    res1 <- rxFastForest(formula = myFormula, data = myXdfDS)
    checkEquals(res1$params$readVars, readVars)
    
    # data frame
    res2 <- rxFastForest(formula = myFormula,
        data = myIris, 
        reportProgress = 4)
    checkEquals(res2$params$readVars, readVars)
    
    res3 <- rxFastForest(formula = myFormula,
        data = myCsvDS) 

    checkEquals(res3$params$readVars, readVars) 
}

"test.rxFastForest.type.regression" <- function()
{
    #  A Simple Regression Forest
    myFormula <- mpg ~ disp
    res1 <- rxFastForest(formula = myFormula, data = mtcars, type = "regression")
    readVars <- c("mpg", "disp")
    checkEquals(res1$params$readVars, readVars)
    
    checkTrue(is.list(coef(res1)))
}

"test.rxFastForest.type.binary" <- function()
{
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.txt")

    dataColInfo <- list(
        RowNum =list(type = "integer"),
        age = list( type = "factor",
            levels = c("17-20", "21-24", "25-29", "30-34", "35-39", "40-49", "50-59", "60+")),
        car.age = list(type = "factor",
            levels = c("0-3", "4-7", "8-9", "10+")),
        type = list(type = "factor",
            levels = c("A", "B", "C", "D")),
		cost = list(type = "float32"),
		number = list(type = "float32"))
	
	dataSource <- RxTextData(dataFile, colInfo = dataColInfo)
	
	type <- "binary"	
	form <- age30 ~ car.age + type + cost + number
	transforms <- quote(list(age30 = (age == "30-34" | age == "35-39")))
	rowSelection <- quote(number > 10)
	writeModelVars <- FALSE
    extraVarsToWrite <- c("RowNum", "age", "age30")
	
    rxSetComputeContext("local")
    model <- rxFastForest(form, data = dataSource, type = type,
	    transforms = transforms, rowSelection = rowSelection
		)
    summ <- summary(model)
    score <- rxPredict(model, data = dataSource, 
		writeModelVars = writeModelVars, extraVarsToWrite = extraVarsToWrite)
    checkEquals(extraVarsToWrite, colnames(score)[seq_along(extraVarsToWrite)])
    
    checkTrue(is.list(coef(model)))
}

"test.rxFastForest.args.general" <- function()
{
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.txt")

    dataColInfo <- list(
        RowNum =list(type = "integer"),
        age = list( type = "factor",
            levels = c("17-20", "21-24", "25-29", "30-34", "35-39", "40-49", "50-59", "60+")),
        car.age = list(type = "factor",
            levels = c("0-3", "4-7", "8-9", "10+")),
        type = list(type = "factor",
            levels = c("A", "B", "C", "D")),
		cost = list(type = "float32"),
		number = list(type = "float32"))
	
	dataSource <- RxTextData(dataFile, colInfo = dataColInfo)
	
	type = "regression"
	numTrees = 100
    numLeaves = 20
    minSplit = 10
    exampleFraction = 0.7
    featureFraction = 0.7
    splitFraction = 0.7
    numBins = 255
    firstUsePenalty = 0
    gainConfLevel = 0
    randomSeed = NULL
    
	form <- cost ~ age + car.age + type + number
	
    rxSetComputeContext("local")
    model <- rxFastForest(form, data = dataSource, type = type,
        numTrees = numTrees, numLeaves = numLeaves, minSplit = minSplit, 
        exampleFraction = exampleFraction, featureFraction = featureFraction, splitFraction = splitFraction, 
        numBins = numBins, firstUsePenalty = firstUsePenalty, gainConfLevel = gainConfLevel, 
        randomSeed = randomSeed
		)
    (summ <- summary(model))
    score <- rxPredict(model, data = dataSource)
}

"test.rxFastForest.args.numTrees" <- function()
{
    #  A Simple Regression Tree    
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number
            
    checkException(
		model1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression", 
            numTrees = 0),
		'Error in processing machine learning request.')
    
    model2 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
        numTrees = 1)  
    out2 <- rxPredict(model2, data = dataFile)      
}

"test.rxFastForest.args.numLeaves" <- function()
{
    #  A Simple Regression Tree
    # FIXME: What is reasonable range?
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number
    res1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
        numLeaves = 10)
    out1 <- rxPredict(res1, data = dataFile)      
}

"test.rxFastForest.args.minSplit" <- function()
{
    #  A Simple Regression Tree
    # FIXME: What is reasonable range?
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number    

    res1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
        minSplit = 10000)
    out1 <- rxPredict(res1, data = dataFile)      
}

"test.rxFastForest.args.exampleFraction" <- function()
{
    #  A Simple Regression Tree
    # FIXME: What is reasonable range?
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number    
    checkException(res1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
        exampleFraction = 10000),
		'Error in processing machine learning request.')
}

"test.rxFastForest.args.featureFraction" <- function()
{
    #  A Simple Regression Tree
    # FIXME: What is reasonable range?
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number
    checkException(res1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
        featureFraction = 10000),
		'Error in processing machine learning request.')
}

"test.rxFastForest.args.splitFraction" <- function()
{
    #  A Simple Regression Tree
    # FIXME: What is reasonable range?
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number
    checkException(res1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
        splitFraction = 10000),
		'Error in processing machine learning request.')
}

"test.rxFastForest.args.numBins" <- function()
{
    #  A Simple Regression Tree
    # FIXME: What is reasonable range?
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number
    res1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
        numBins = 10000)
    out1 <- rxPredict(res1, data = dataFile)      
}

"test.rxFastForest.args.firstUsePenalty" <- function()
{
    #  A Simple Regression Tree
    # FIXME: What is reasonable range?
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number
    res1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
        firstUsePenalty = 10000)
    out1 <- rxPredict(res1, data = dataFile)      
}

"test.rxFastForest.args.gainConfLevel" <- function()
{
    #  A Simple Regression Tree
    # Range: [0,1)
       
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number
    res1 <-
        rxFastForest(formula = myFormula, data = dataFile, type = "regression",
            gainConfLevel = 0)
    
    out1 <- rxPredict(res1, data = dataFile)  
    
    # This should fail       
    res2 <- try(
        rxFastForest(formula = myFormula, data = dataFile, type = "regression",
            gainConfLevel = 1), silent = TRUE)
        
    checkTrue(inherits( res2, "try-error")) 
    checkTrue(grepl("gainConfLevel' must be less than 1", as.character(res2)))      
}

"test.rxFastForest.args.trainThreads" <- function()
{
    # Simple binary tree
    myIris <- iris
    myIris$isVersi <- iris$Species == "versicolor"
    irisFormula <- isVersi~Sepal.Length + Sepal.Width + Petal.Length + Petal.Width
    tm1 <- system.time(
        checkException(res1 <- rxFastForest(formula = irisFormula, data = myIris, 
            trainThreads = -1, reportProgress = 0),
			'Error in processing machine learning request.') )   
        
    tm2 <- system.time(
        res2 <- rxFastForest(formula = irisFormula, data = myIris,
           trainThreads = 4, reportProgress = 0) )              
}

"test.rxFastForest.args.verbose" <- function()
{
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number    
    # This should give no output
    out1 <- capture.output(
        res1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
           verbose = 0, numTrees = 2, reportProgress = 0)
    )

    checkTrue(if (rxIsExpressEdition()) length(out1) > 0 else length(out1) == 0)
    
    # This should give  output
    out3 <- capture.output(
        res3 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
           verbose = 1, reportProgress = 0)
    )

    checkTrue(length(out3) > 0)    
}

"test.rxFastForest.args.hidden" <- function()
{
    #  A Simple Regression Tree
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
	myFormula <- cost ~ age + car.age + type + number    
        
    res1 <- rxFastForest(formula = myFormula, data = dataFile, type = "regression",
        exampleFraction = 0.7,
        quantileSampleCount = 100,
        diskTranspose = FALSE,
        sparsifyThreshold = 0.7,
        gainConfLevel = 0,
        softmaxTemperature = 0,
        executionTimes = FALSE)
    
    checkEquals(res1$params$readVars, all.vars(myFormula)) 
    checkTrue(grepl("quantileSampleCount=100", res1$mamlCode)) 
    checkTrue(grepl("executionTimes=-", res1$mamlCode)) 

}

"test.rxFastForest.data.integer" <- function()
{
DEACTIVATED("coefficients property needs to be enabled")
    inData <- file.path(rxGetOption("sampleDataDir"), "claims4blocks.xdf")
    rxGetVarInfo(inData)
    modelOut1 <- rxFastForest(RowNum ~ age + cost, data = inData, type = "regression")
	sum1 <- summary(modelOut1)
	
	inDF <- rxDataStep(inData)
	inDF$RowNum <- as.numeric(inDF$RowNum)
    modelOut2 <- rxFastForest(RowNum ~ age + cost, data = inDF, type = "regression")
    sum2 <- summary(modelOut2)

    # Check for non nulls
    checkTrue(!is.null(sum1$coefficients))
    checkTrue(!is.null(sum2$coefficients))

	# Check to make sure the results are the same
	checkEquals(sum1$coefficients, sum2$coefficients)
	
}

"test.rxFastForest.MAML.code.check.optional.params" <- function() 
{
    # Create temporary data sources
    myIris = iris
    myIris$Setosa <- iris$Species == "setosa"
    defaultParameters <- "lab={Setosa} feat={Sepal.Length} feat={Petal.Length} feat={Sepal.Width}   normalize=Auto trainer=FastForestClassification{ featureFraction=0.7 baggingTrainFraction=0.7 splitFraction=0.7 numThreads=8 rngSeed=123 entropyCoefficient=0 histogramPoolSize=-1 diskTranspose=- maxBins=255 sparsifyThreshold=0.7 featureFirstUsePenalty=0 featureReusePenalty=0 gainConfidenceLevel=0 softmaxTemperature=0 executionTimes=- numLeaves=20 minDocumentsInLeafs=10 numTrees=100 smoothing=0 testFrequency=2147483647 baggingSize=1} calibrator=sigmoid"
    allParamsChanged <- "lab={Setosa} feat={Sepal.Length} feat={Petal.Length} feat={Sepal.Width}   normalize=Auto trainer=FastForestClassification{ featureFraction=0.8 baggingTrainFraction=0.8 splitFraction=0.8 numThreads=2 rngSeed=5 entropyCoefficient=0 histogramPoolSize=-1 diskTranspose=- maxBins=254 sparsifyThreshold=0.7 featureFirstUsePenalty=1 featureReusePenalty=0 gainConfidenceLevel=0.5 softmaxTemperature=0 executionTimes=- numLeaves=23 minDocumentsInLeafs=11 numTrees=90 smoothing=0 testFrequency=2147483647 baggingSize=1} calibrator=sigmoid randomSeed=5"
    
    # Run examples
    myFormula <- Setosa ~ Sepal.Length + Petal.Length + Sepal.Width
       
    resDefault <- rxFastForest(formula = myFormula, data = myIris)
    resAllParamsChanged <- rxFastForest(formula = myFormula, data = myIris, trainThreads = 2, type = "binary", numTrees = 90, numLeaves = 23, minSplit = 11, exampleFraction = 0.8, featureFraction = 0.8, splitFraction = 0.8, numBins = 254, firstUsePenalty = 1, gainConfLevel = 0.5, randomSeed = 5)
    checkEquals(resDefault$mamlCode, defaultParameters)
    checkEquals(resAllParamsChanged$mamlCode, allParamsChanged)
}

"test.rxFastForest.ensemble.binary" <- function()
{
    #library(MicrosoftML)
    #rm(list=ls())

    myDf <- within(iris, {Setosa <- (Species == "setosa")})
    myFormula <- Setosa ~ Sepal.Length + Sepal.Width + Petal.Length + Sepal.Width + Species

    randomSeed <- 111
    replace <- TRUE
    sampRate <- 0.75
    combineMethod <- NULL

    # data frame
    modelCount <- 1
    resDf <- rxFastForest(formula = myFormula, data = myDf
        , ensemble = ensembleControl(modelCount = modelCount, replace = replace)
        )
    checkTrue(!inherits(resDf, "Ensemble"))
    #print(structure(resDf$mamlCode, class = "maml"))

    # xdf file
    myXdf <- tempfile(pattern = "myXdf", fileext = ".xdf")
    on.exit(unlink(myXdf), add = TRUE)
    rxDataStep(myDf, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)

    modelCount <- 2
    resXdf <- rxFastForest(formula = myFormula, data = myXdfDS, randomSeed = NULL
        , ensemble = ensembleControl(modelCount = modelCount, sampRate = sampRate, randomSeed = randomSeed)
        )
    checkTrue(inherits(resXdf, "Ensemble"))
    #cat(resXdf$mamlCode, "\n")

    # csv
    myCsv <- tempfile(pattern = "myCsv", fileext = ".csv")
    on.exit(unlink(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, colInfo = rxCreateColInfo(iris))
    rxDataStep(myDf, myCsvDS, overwrite = TRUE)

    modelCount <- 3
    resCsv <- rxFastForest(formula = myFormula, data = myCsvDS, randomSeed = 222
        , ensemble = ensembleControl(modelCount = modelCount, combineMethod = combineMethod, randomSeed = randomSeed)
        )
    checkTrue(inherits(resCsv, "Ensemble"))
    #cat(resCsv$mamlCode, "\n")
}

"test.rxFastForest.ensemble.regression" <- function()
{
    #library(MicrosoftML)
    #rm(list=ls())

    myDf <- within(iris, {Setosa <- (Species == "setosa")})
    myFormula <- Sepal.Length ~ Sepal.Width + Petal.Length + Sepal.Width + Species

    randomSeed <- 111
    replace <- TRUE
    sampRate <- 0.75
    combineMethod <- NULL
    type <- "regression"

    # data frame
    modelCount <- 1
    resDf <- rxFastForest(formula = myFormula, data = myDf, type = type
        , ensemble = ensembleControl(modelCount = modelCount, replace = replace)
        )
    checkTrue(!inherits(resDf, "Ensemble"))
    #print(structure(resDf$mamlCode, class = "maml"))

    # xdf file
    myXdf <- tempfile(pattern = "myXdf", fileext = ".xdf")
    on.exit(unlink(myXdf), add = TRUE)
    rxDataStep(myDf, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)

    modelCount <- 2
    resXdf <- rxFastForest(formula = myFormula, data = myXdfDS, randomSeed = NULL, type = type
        , ensemble = ensembleControl(modelCount = modelCount, sampRate = sampRate, randomSeed = randomSeed)
        )
    checkTrue(inherits(resXdf, "Ensemble"))
    #cat(resXdf$mamlCode, "\n")

    # csv
    myCsv <- tempfile(pattern = "myCsv", fileext = ".csv")
    on.exit(unlink(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, colInfo = rxCreateColInfo(iris))
    rxDataStep(myDf, myCsvDS, overwrite = TRUE)

    modelCount <- 3
    resCsv <- rxFastForest(formula = myFormula, data = myCsvDS, randomSeed = 222, type = type
        , ensemble = ensembleControl(modelCount = modelCount, combineMethod = combineMethod, randomSeed = randomSeed)
        )
    checkTrue(inherits(resCsv, "Ensemble"))
    #cat(resCsv$mamlCode, "\n")
}
