# Copyright (c) Microsoft Corporation.  All rights reserved.

# Tests for basic functionality of rxFastLinear
".rxSqliteConnectionStringFromFile" <- function() {
    if (exists("rxSqliteConnectionStringFromFile", globalenv()))
        get("rxSqliteConnectionStringFromFile", globalenv())
    else
        RevoScaleR:::rxSqliteConnectionStringFromFile
}

"test.rxFastLinear.basic.binary" <- function()
{
    # Create temporary data sources
    myIris = iris
    myIris$Setosa <- iris$Species == "setosa"
    
    myXdf <- tempfile(pattern = "myIris", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)
    rxDataStep(myIris, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)
    
    myCsv <- tempfile(pattern = "myIris", fileext = ".csv")
    on.exit(file.remove(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, rowsPerRead = 40)
    rxDataStep(myIris, myCsvDS, overwrite = TRUE)
    
    readVars <- c("Setosa", "Sepal.Length", "Sepal.Width",  "Petal.Width") 
    # Run examples
    res1 <- rxFastLinear(formula = Setosa ~ Sepal.Length + Sepal.Width + Petal.Width,
        data = myXdfDS, type = "binary", shuffle = FALSE)
    checkEquals(res1$params$readVars, readVars)
	scoreOut <- rxPredict(res1, data = myXdfDS, extraVarsToWrite = "Setosa")
    checkTrue(is.vector(coef(res1)))
    checkEquals(coef(res1), coef(summary(res1)))
    checkEquals(res1$coefficients, coef(res1))
    sumRes1 <- summary(res1, top = 2)
    checkEquals(length(coef(sumRes1)), 2)
    checkEquals(names(coef(sumRes1))[1], "(Bias)")
       
    res2 <- rxFastLinear(formula = Setosa ~ Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris, type = "binary", shuffle = FALSE)
    checkEquals(res2$params$readVars, readVars)
    
    res3 <- rxFastLinear(formula = Setosa ~ Sepal.Length + Sepal.Width + Petal.Width,
        data = myCsvDS, type = "binary", shuffle = FALSE)
    checkEquals(res3$params$readVars, readVars)
}

"test.rxFastLinear.basic.regression" <- function()
{
	summary(fm1 <- lm(rating ~ ., data = attitude))

    # Create temporary data sources
    myXdf <- tempfile(pattern = "tempAttitude", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)
    rxDataStep(attitude, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)
    
    myCsv <- tempfile(pattern = "tempAttitude", fileext = ".csv")
    on.exit(file.remove(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, rowsPerRead = 40)
    rxDataStep(attitude, myCsvDS, overwrite = TRUE)
    
	attitudeForm <- rating~complaints+privileges+learning+raises+critical+advance

	# Xdf example
    res1 <- rxFastLinear(formula = attitudeForm, data = myXdfDS, type = "regression", 
        l2Weight = 0.01, l1Weight = 0, maxIterations = 20, trainThreads = 1, shuffle = FALSE)
    scoreOut1 <- rxPredict(res1, data = myXdfDS, extraVarsToWrite = "rating")
    checkTrue(is.vector(coef(res1)))
    checkEquals(coef(res1), coef(summary(res1)))
    checkEquals(res1$coefficients, coef(res1))
    sumRes1 <- summary(res1, top = 2)
    checkEquals(length(coef(sumRes1)), 2)
    checkEquals(names(coef(sumRes1))[1], "(Bias)")
       
    res2 <- rxFastLinear(formula = attitudeForm, type = "regression",
        data = attitude, l2Weight = 0.01, l1Weight = 0, maxIterations = 20, trainThreads = 1, shuffle = FALSE)
    scoreOut2 <- rxPredict(res2, data = attitude, extraVarsToWrite = "rating")
    checkEquals(coef(res1), coef(res2))
    checkEquals(scoreOut1, scoreOut2)
	
    res3 <- rxFastLinear(formula = attitudeForm, type = "regression",
        data = myCsvDS, l2Weight = 0.01, l1Weight = 0, maxIterations = 20, trainThreads = 1, shuffle = FALSE)
    scoreOut3 <- rxPredict(res2, data = myCsvDS, extraVarsToWrite = "rating")
    checkEquals(coef(res1), coef(res3))
    checkEquals(scoreOut1, scoreOut3)
}

"test.rxFastLinear.args.binary.rxLogisticRegression" <- function()
{
    infert1 <- infert
    infert1$isCase <- (infert1$case == 1)
    
    type <- "binary"
    l2Weight <- 1e-6
    l1Weight <- 1e-6
    optTol <- 1e-07
    maxIterations <- 10 #1e5
    normalize <- "auto"

    form <- isCase ~ age + parity + education + spontaneous + induced
    
    # Estimate a logistic regression model
    system.time(
    logitModel <- rxLogisticRegression(form, data = infert1, type = type
        , l2Weight = l2Weight, l1Weight = l1Weight, normalize = normalize
        , optTol = optTol, maxIterations = maxIterations
        #, trainThreads = NULL        
        #, memorySize = 20
        #, initWtsScale = 0
        #, showTrainingStats = FALSE
        #, sgdInitTol = 0
        #, denseOptimizer = FALSE
        )
    )
    coef(logitModel)
    
    # Estimate a linear regression model
    system.time(
    linearModel <- rxFastLinear(form, data = infert1, type = type
        , l2Weight = l2Weight, l1Weight = l1Weight, normalize = normalize
        , convergenceTolerance = optTol, maxIterations = maxIterations
        #, trainThreads = NULL
        #, lossFunction = NULL
        , shuffle = FALSE
        #, checkFrequency = NULL
        )
    )
    coef(linearModel)
    
    ##TODO: can we expect that they give the same coefs and rxFastLinear is indeed faster?
    checkTrue(TRUE)
}

"test.rxFastLinear.mlTransforms.mtText.punctuations" <- function() 
{
    mytext <- c(
        "!", "\"", "#", "$", "%", "&", "'", "(", ")", "*", "+", ",", "-", ".", "/", ":", 
        ";", "<", "=", ">", "?", "@", "[", "\\", "]", "^", "_", "`", "{", "|", "}", "~")
    myrating <- rep_len(0:1, length(mytext))
    mydata <- data.frame(Rating = myrating, Text = mytext, stringsAsFactors = FALSE)

    mymodel1 <- rxFastLinear(formula = Rating ~ TextFeaturized, data = mydata, #trainThreads = 1,
        mlTransforms = list(featurizeText(vars = c(TextFeaturized = "Text"))), shuffle = FALSE)
    mymodel2 <- rxFastLinear(formula = Rating ~ TextFeaturized, data = mydata, #trainThreads = 1,
        mlTransforms = list(featurizeText(vars = c(TextFeaturized = "Text"), 
        keepPunctuations = TRUE)), shuffle = FALSE)

    mycoef <- union(names(coef(mymodel1)), names(coef(mymodel2)))
    #length(mycoef)
    #checkTrue(all(c(mytext, "(Bias)") %in% mycoef))
    checkTrue(all(mycoef %in% c(mytext, "(Bias)")))
}

"test.rxFastLinear.reproducibility" <- function()
{
    myIris = iris
    myIris$Setosa <- iris$Species == "setosa"

    model0 <- rxFastLinear(formula = Setosa ~ Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris, type = "binary", trainThreads = 1, shuffle = FALSE)

    for (i in 1:10) {
        model <- rxFastLinear(formula = Setosa ~ Sepal.Length + Sepal.Width + Petal.Width,
            data = myIris, type = "binary", trainThreads = 1, shuffle = FALSE)
        checkEquals(coef(model0), coef(model))
    }
}

"test.rxFastLinear.MAML.code.check.optional.params" <- function() 
{
    myIris = iris
    myIris$Setosa <- iris$Species == "setosa"
    defaultMamlCode <- "lab={Setosa} feat={Sepal.Length} feat={Sepal.Width} feat={Petal.Width}   normalize=Auto trainer=SDCA{ lossFunction=LogLoss convergenceTolerance=0.1 shuffle=- positiveInstanceWeight=1} calibrator=sigmoid postTransformCache=Disk"
    allParamsChanged <- "lab={Setosa} feat={Sepal.Length} feat={Sepal.Width} feat={Petal.Width}   normalize=warn trainer=SDCA{ lossFunction=LogLoss l2Const=1 l1Threshold=1 numThreads=1 convergenceTolerance=0.2 maxIterations=50 shuffle=+ checkFrequency=1 positiveInstanceWeight=1} calibrator=sigmoid postTransformCache=Disk"
            
    # Run examples
    resDefault <- rxFastLinear(formula = Setosa ~ Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris, type = "binary", shuffle = FALSE)
    resAllParamsChanged <- rxFastLinear(formula = Setosa ~ Sepal.Length + Sepal.Width + Petal.Width, 
        data = myIris, type = "binary", l1Weight = 1, l2Weight = 1, trainThreads = 1, convergenceTolerance = 0.2, 
        maxIterations = 50, shuffle = TRUE, checkFrequency = 1, normalize = "warn", verbose = 2)

    checkEquals(resAllParamsChanged$mamlCode, allParamsChanged)
    checkEquals(resDefault$mamlCode, defaultMamlCode)
}

"test.rxFastLinear.ensemble.binary" <- function()
{
    #library(MicrosoftML)
    #rm(list=ls())

    myDf <- within(iris, {Setosa <- (Species == "setosa")})
    myFormula <- Setosa ~ Sepal.Length + Sepal.Width + Petal.Length + Sepal.Width + Species

    randomSeed <- 111
    replace <- TRUE
    sampRate <- 0.75
    combineMethod <- NULL

    # data frame
    modelCount <- 1
    resDf <- rxFastLinear(formula = myFormula, data = myDf
        , ensemble = ensembleControl(modelCount = modelCount, replace = replace)
        , shuffle = FALSE)
    checkTrue(!inherits(resDf, "Ensemble"))
    #print(structure(resDf$mamlCode, class = "maml"))

    # xdf file
    myXdf <- tempfile(pattern = "myXdf", fileext = ".xdf")
    on.exit(unlink(myXdf), add = TRUE)
    rxDataStep(myDf, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)

    modelCount <- 2
    resXdf <- rxFastLinear(formula = myFormula, data = myXdfDS
        , ensemble = ensembleControl(modelCount = modelCount, sampRate = sampRate)
        , shuffle = FALSE)
    checkTrue(inherits(resXdf, "Ensemble"))
    #cat(resXdf$mamlCode, "\n")

    # csv
    myCsv <- tempfile(pattern = "myCsv", fileext = ".csv")
    on.exit(unlink(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, colInfo = rxCreateColInfo(iris))
    rxDataStep(myDf, myCsvDS, overwrite = TRUE)

    modelCount <- 3
    resCsv <- rxFastLinear(formula = myFormula, data = myCsvDS
        , ensemble = ensembleControl(modelCount = modelCount, combineMethod = combineMethod, randomSeed = randomSeed)
        , shuffle = FALSE)
    checkTrue(inherits(resCsv, "Ensemble"))
    #cat(resCsv$mamlCode, "\n")
}

"test.rxFastLinear.ensemble.regression" <- function()
{
    #library(MicrosoftML)
    #rm(list=ls())

    myDf <- within(iris, {Setosa <- (Species == "setosa")})
    myFormula <- Sepal.Length ~ Sepal.Width + Petal.Length + Sepal.Width + Species

    randomSeed <- 111
    replace <- TRUE
    sampRate <- 0.75
    combineMethod <- NULL
    type <- "regression"

    # data frame
    modelCount <- 1
    resDf <- rxFastLinear(formula = myFormula, data = myDf, type = type
        , ensemble = ensembleControl(modelCount = modelCount, replace = replace)
        , shuffle = FALSE)
    checkTrue(!inherits(resDf, "Ensemble"))
    #print(structure(resDf$mamlCode, class = "maml"))

    # xdf file
    myXdf <- tempfile(pattern = "myXdf", fileext = ".xdf")
    on.exit(unlink(myXdf), add = TRUE)
    rxDataStep(myDf, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)

    modelCount <- 2
    resXdf <- rxFastLinear(formula = myFormula, data = myXdfDS, type = type
        , ensemble = ensembleControl(modelCount = modelCount, sampRate = sampRate, randomSeed = randomSeed)
        , shuffle = FALSE)
    checkTrue(inherits(resXdf, "Ensemble"))
    #cat(resXdf$mamlCode, "\n")

    # csv
    myCsv <- tempfile(pattern = "myCsv", fileext = ".csv")
    on.exit(unlink(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, colInfo = rxCreateColInfo(iris))
    rxDataStep(myDf, myCsvDS, overwrite = TRUE)

    modelCount <- 3
    resCsv <- rxFastLinear(formula = myFormula, data = myCsvDS, type = type
        , ensemble = ensembleControl(modelCount = modelCount, combineMethod = combineMethod, randomSeed = randomSeed)
        , shuffle = FALSE)
    checkTrue(inherits(resCsv, "Ensemble"))
    #cat(resCsv$mamlCode, "\n")
}
