# Copyright (c) Microsoft Corporation.  All rights reserved.

# Tests for basic functionality of rxLogisticRegression
".rxSqliteConnectionStringFromFile" <- function()
{
    if (exists("rxSqliteConnectionStringFromFile", globalenv()))
        get("rxSqliteConnectionStringFromFile", globalenv())
    else
        RevoScaleR:::rxSqliteConnectionStringFromFile
}

"test.rxLogisticRegression.basic" <- function()
{
    # Create temporary data sources
    myIris = iris
    myIris$IsVersi <- iris$Species == "versicolor"
    
   
    # Create a temporary multi-block xdf file
    myXdf <- tempfile(pattern = "myIris", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)
    rxDataStep(myIris, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)
    
    # Create a temporary csv file
    myCsv <- tempfile(pattern = "myIris", fileext = ".csv")
    on.exit(file.remove(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, rowsPerRead = 40)
    rxDataStep(myIris, myCsvDS, overwrite = TRUE)
    
    # Xdf file
    irisForm <- IsVersi~Sepal.Length + Sepal.Width + Petal.Length + Petal.Width
    res1 <- rxLogisticRegression(formula = irisForm,
        data = myXdfDS)
    sumXdf <- summary(res1)
    checkTrue(is.vector(coef(res1)))
    checkEquals(coef(res1), coef(sumXdf))
    checkEquals(res1$coefficients, coef(res1))
    
	# Data frame      
    res2 <- rxLogisticRegression(formula = irisForm,
        data = myIris)
    sumDF <- summary(res2)
    checkEquals(coef(sumXdf), coef(sumDF))
    
    # CSV
    res3 <- rxLogisticRegression(formula = irisForm,
        data = myCsvDS)
    sumCsv <- summary(res3)
    checkEquals(coef(sumXdf), coef(sumCsv))

    sumCsv2 <- summary(res3, top = 2)
    checkEquals(length(coef(sumCsv2)), 2)
    checkEquals(names(coef(sumCsv2))[1], "(Bias)")

    resStats <- rxLogisticRegression(formula = irisForm,
        data = myCsvDS, showTrainingStats = TRUE)
    sumStats <- summary(resStats)
    checkEquals(coef(sumXdf), coef(sumStats))
    sumStatsTop2 <- summary(resStats, top = 2)
    checkEquals(length(coef(sumStatsTop2)), 2)
    checkEquals(names(coef(sumStatsTop2))[1], "(Bias)")
}

"test.rxLogisticRegression.chunking" <- function()
{
  # FIXME: What is expected behavior for chunking?
    myIris = iris
    myIris$Setosa <- iris$Species == "setosa"
    
    myXdf <- tempfile(pattern = "myIris", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)

    rxDataStep(myIris, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)

    res1 <- rxLogisticRegression(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        data = myXdfDS)
    
    myXdfOut <- tempfile(pattern = "myIrisOut", fileext = ".xdf")
    on.exit(file.remove(myXdfOut), add = TRUE)
        
    rxPredict(res1, data = myXdf, outData = myXdfOut, reportProgress = 1)
    
}

"test.rxLogisticRegression.args.transforms" <- function()
{
    # Create temporary data sources
    myIris = iris

    myXdf <- tempfile(pattern = "myIris", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)
    rxDataStep(myIris, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)
    
    myCsv <- tempfile(pattern = "myIris", fileext = ".csv")
    on.exit(file.remove(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, rowsPerRead = 50)
    rxDataStep(myIris, myCsvDS, overwrite = TRUE)
    
    readVars <- c("Setosa", "Sepal.Length", "Sepal.Width",  "Petal.Width",  "Species")  
    # Run examples
    res1 <- rxLogisticRegression(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        transforms = list(Setosa = Species == "setosa"), transformVars = c("Species"),
        data = myXdfDS)
    checkEquals(res1$params$readVars, readVars)
    
    res2 <- rxLogisticRegression(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        transforms = list(Setosa= Species == "setosa"), transformVars = "Species",
        data = iris)
    checkEquals(res2$params$readVars, readVars)
    
    res3 <- rxLogisticRegression(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        data = myCsvDS, 
        transforms = list(Setosa = Species == "setosa"), transformVars = "Species")
    checkEquals(res3$params$readVars, readVars)

    # Verify cache options propogates properly
    res4 <- rxLogisticRegression(formula = Setosa ~ Sepal.Length + Sepal.Width + Petal.Width,
        data = myCsvDS,
        transforms = list(Setosa = Species == "setosa"), transformVars = "Species", preTransformCache = "Disk", postTransformCache = "Memory")
    checkTrue(grepl("preTransformCache=Disk", res4$mamlCode))
    checkTrue(grepl("postTransformCache=Memory", res4$mamlCode))
}

"test.rxLogisticRegression.args.transformFunc" <- function()
{
    textOutFile <- tempfile(pattern = "logitScore", fileext = ".txt")
    myFormula <- isCase ~ age + parity + education + spontaneous + induced
    
    ######################################################################################
    # Model without transforms
    ######################################################################################
    infert1 <- infert
    infert1$isCase = infert1$case == 1
    infert1$case <- NULL
    
    # Estimate a logistic regression model 
    logitModel <- rxLogisticRegression(formula = myFormula,
                           data = infert1)                   

    # Put the per-instance results in a text file
    scoreDS <- rxPredict( logitModel, 
         data = infert1, extraVarsToWrite = c("isCase", "Probability"),
         outData = textOutFile, overwrite = TRUE)
    dfWithoutTransforms = rxDataStep(scoreDS)
    file.remove(textOutFile)  
        
    ######################################################################################
    # Model with transforms
    ######################################################################################      

    # Estimate a logistic regression model 
    logitModel <- rxLogisticRegression(formula = myFormula,
                           data = infert, transforms = list(isCase = case == 1))
                       
    # Put the per-instance results in a text file
    scoreDS <- rxPredict(logitModel, 
         data = infert, extraVarsToWrite = c("isCase", "Probability"),
         outData = textOutFile)
    dfWithTransforms = rxDataStep(scoreDS)
    file.remove(textOutFile)
    
    checkIdentical(dfWithoutTransforms, dfWithTransforms)
    
    ######################################################################################
    # Model with transformFunc
    ######################################################################################      

    myTransforms <- function(dataList)
    {
        dataList$isCase = dataList$case == 1
        return( dataList )
    }
    
    # Estimate a logistic regression model 
    logitObject <- rxLogisticRegression(formula = myFormula,
                           data = infert, transformFunc = myTransforms,
                           transformVars = "case")                           
                       

    # Put the per-instance results in a text file
    scoreDS <- rxPredict(logitObject, 
         data = infert, extraVarsToWrite = c("isCase", "Probability"),
         outData = textOutFile)
    dfWithFunc = rxDataStep(scoreDS)
    file.remove(textOutFile)
    
    checkIdentical(dfWithoutTransforms, dfWithFunc)    
    
    ######################################################################################
    # Formula with transform
    ######################################################################################      

    # Estimate a logistic regression model 
    logitModel <- rxLogisticRegression(formula = (case == 1) ~ age + parity + education + spontaneous + induced,
                           data = infert)
                       

    # Put the per-instance results in a data frame
    dfWithForm <- rxPredict(logitModel, 
         data = infert, extraVarsToWrite = c("X.case....1.", "Probability"),
         outData = NULL)

    namesWithForm <- names(dfWithForm)
    namesWithForm[1] <- "isCase"
    names(dfWithForm) <- namesWithForm
    # Note Score and Probability are different precision in the two cases
    checkEquals(dfWithoutTransforms, dfWithForm, tolerance = 1E-06)        
    
}
"test.rxLogisticRegression.args.mlTransforms" <- function()
{
    myFormula <- isCase ~ age + parity + education + spontaneous + induced
    
    ######################################################################################
    # Model without transforms
    ######################################################################################
    infert1 <- infert
    infert1$isCase = infert1$case == 1
    infert1$case <- NULL
    
    # Estimate a logistic regression model 
    logitModel <- rxLogisticRegression(formula = myFormula,
                           data = infert1)                   

    # Score the model
    dfWithoutTransforms <- rxPredict( logitModel, 
         data = infert1, extraVarsToWrite = c("isCase", "Probability"),
         outData = NULL, overwrite = TRUE)
    
    ######################################################################################
    # Model with MAML transforms
    ######################################################################################      

    # Estimate a logistic regression model 
    logitModel <- rxLogisticRegression(formula = myFormula,
                           data = infert,
						   mlTransforms = list("Expr{col=isCase:case expr={x : x == 1}} "),
						   mlTransformVars = "case")
                       
    dfWithTransforms <- rxPredict(logitModel, 
         data = infert, extraVarsToWrite = c("isCase", "Probability"),
         outData = NULL)
    
    checkIdentical(dfWithoutTransforms, dfWithTransforms)    
}

"test.rxLogisticRegression.args.reportProgress" <- function()
{
    # Create temporary data sources
    DEACTIVATED("Deactivated until report progress output is finalized.")
    myIris = iris
    myIris$Setosa <- iris$Species == "setosa"
    
    myXdf <- tempfile(pattern = "myIris", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)
    rxDataStep(myIris, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)
    
    myCsv <- tempfile(pattern = "myIris", fileext = ".csv")
    on.exit(file.remove(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, rowsPerRead = 25)
    rxDataStep(myIris, myCsvDS, overwrite = TRUE)
    
    readVars <- c("Setosa", "Sepal.Length", "Sepal.Width",  "Petal.Width") 
    # Run examples
    # Set trainThreads to 1 to turn of warning
    out0 <- capture.output(res0 <- rxLogisticRegression(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        data = myXdfDS, trainThreads = 1, verbose = 0, reportProgress = 0))
    checkEquals(length(out0), 0)
    
    out1 <- capture.output(res1 <- rxLogisticRegression(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        data = myXdfDS, trainThreads = 1, verbose = 0, reportProgress = 1))
    checkTrue(length(out1) >= 208)    
    
    out2 <- capture.output(res2 <- rxLogisticRegression(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        data = myXdfDS, trainThreads = 1, verbose = 0, reportProgress = 2))
    checkTrue(length(out2) >= 208)
    CheckStringsFound(out2, "Read Time")
        
    # Text data
    out3 <- capture.output(res3 <- rxLogisticRegression(formula = Setosa~Sepal.Length + Sepal.Width + Petal.Width,
        data = myCsvDS, trainThreads = 1, verbose = 0, reportProgress = 2))
    checkTrue(length(out3) >= 364) 
    CheckStringsFound(out3, "Rows Read: 25")
}


"test.rxLogisticRegression.args.data.claims" <- function()
{
    ###############################################################################################
    # Xdf - One Block
    claimsXdf <- file.path(rxGetOption("sampleDataDir"), "claims.xdf")
    lregOut <- rxLogisticRegression(car.age == "8-9"~ number + cost + age, data = claimsXdf)
    sumXdf <- summary(lregOut)
                          
    ##############################################################################################
    # Data frame
    claimsDF <- rxDataStep(inData = claimsXdf)
    lregOut <- rxLogisticRegression(car.age == "8-9"~ number + cost + age, data = claimsDF)
    sumDF <- summary(lregOut)    

    checkEquals(coef(sumXdf), coef(sumDF))
    ##############################################################################################
    # Multi-block XDF
    claimsXdfMulti <- file.path(rxGetOption("sampleDataDir"), "claims4blocks.xdf")
    lregOut <- rxLogisticRegression(car.age == "8-9"~ number + cost + age, data = claimsXdfMulti)
    sumXdfMulti <- summary(lregOut)
    checkEquals(coef(sumXdf), coef(sumXdfMulti))    
    
    ##############################################################################################
    # Composite XDF
    claimsXdfComposite <- file.path(rxGetOption("sampleDataDir"), "claimsXdf")
    lregOut <- rxLogisticRegression(car.age == "8-9"~ number + cost + age, data = claimsXdfComposite)
    sumXdfComposite <- summary(lregOut)
    checkEquals(coef(sumXdf), coef(sumXdfComposite))        
    
    ##############################################################################################
    # SAS
    myColInfo = rxCreateColInfo(claimsDF)
    claimsFileSAS <- file.path(rxGetOption("sampleDataDir"), "claims.sas7bdat")
    claimsSAS <- RxSasData(file = claimsFileSAS, colInfo = myColInfo, rowsPerRead = 50)
    lregOut <- rxLogisticRegression(car_age == "8-9"~ number + cost + age, data = claimsSAS)
    sumSAS <- summary(lregOut)
    checkEquals(coef(sumXdf), coef(sumSAS))           
    
    ##############################################################################################
    # SPSS
    claimsFileSPSS <- file.path(rxGetOption("sampleDataDir"), "claims.sav")
    claimsSPSS <- RxSpssData(file = claimsFileSPSS, colInfo = myColInfo)
    lregOut <- rxLogisticRegression(car_age == "8-9"~ number + cost + age, data = claimsSPSS)
    sumSPSS <- summary(lregOut)  
    checkEquals(coef(sumXdf), coef(sumSPSS))        
    
   ##############################################################################################
    # Text
    claimsFileTxt <- file.path(rxGetOption("sampleDataDir"), "claims.txt")
    claimsTxt <- RxTextData(file = claimsFileTxt, colInfo = myColInfo, rowsPerRead = 100, useFastRead = TRUE)
    lregOut <- rxLogisticRegression(car.age == "8-9"~ number + cost + age, data = claimsTxt)
    sumText1 <- summary(lregOut) 
    checkEquals(coef(sumXdf), coef(sumText1))      
    
    claimsTxt2 <- RxTextData(file = claimsFileTxt, colInfo = myColInfo, rowsPerRead = 100, useFastRead = FALSE)
    lregOut <- rxLogisticRegression(car.age == "8-9"~ number + cost + age, data = claimsTxt2)  
    sumText2 <- summary(lregOut) 
    checkEquals(coef(sumXdf), coef(sumText2))  
        
   ##############################################################################################    
    # ODBC
    rxSqliteConnectionStringFromFile <- .rxSqliteConnectionStringFromFile()
    conString <-
      try(rxSqliteConnectionStringFromFile(file.path(rxGetOption("sampleDataDir"),
                                                    "claims.sqlite")),
          silent = TRUE) 
    
    # Deactivate if Sqlite Driver not available
    if (inherits(conString, "try-error"))
        DEACTIVATED(strsplit(conString, "\n")[[1]][2])
        
    claimsODBC <-  RxOdbcData(sqlQuery = "SELECT * FROM claims",
                 connectionString = conString, colInfo = myColInfo)
    lregOut <- rxLogisticRegression(car_age == "8-9"~ number + cost + age, data = claimsODBC)  
    sumODBC1 <- summary(lregOut) 
    checkEquals(coef(sumXdf), coef(sumODBC1))   
    
    claimsODBC2 <-  RxOdbcData(table = "claims",
                 connectionString = conString, colInfo = myColInfo)
    lregOut <- rxLogisticRegression(car_age == "8-9"~ number + cost + age, data = claimsODBC2)  
    sumODBC2 <- summary(lregOut) 
    checkEquals(coef(sumXdf), coef(sumODBC2))               
 
 }

"test.rxLogisticRegression.args.data.dataframe" <- function()
{
    #DEACTIVATED("Bug 21290: Customer Issue: using data frame as input bloats serialized model size.")

    form <- case ~ age + parity + education + spontaneous + induced
    
    ## run the model for the first time
    model.lr1 <- rxLogisticRegression(form, data = infert)
    model.bin1 <- serialize(model.lr1, connection = NULL)
    model.gz1 <- memCompress(model.bin1, type = "gzip")
    (size1 <- object.size(model.gz1))
    
    ## run the same model again
    model.lr2 <- rxLogisticRegression(form, data = infert)
    model.bin2 <- serialize(model.lr2, connection = NULL)
    model.gz2 <- memCompress(model.bin2, type = "gzip")
    (size2 <- object.size(model.gz2))

    ## make sure no extra stuff in the global env are included in the model
    checkTrue(abs(size2 - size1) < size1 * 0.1)
}

"test.rxLogisticRegression.args.l1l2.ExpressEdition" <- function() 
{
    utils.run.ExpressEdition(run.rxLogisticRegression.args.l1l2.ExpressEdition)
}

"run.rxLogisticRegression.args.l1l2.ExpressEdition" <- function()
{
    # Verifies that we are really running under Express Edition.
    checkTrue(rxIsExpressEdition())

    myInfert <- infert
    myInfert$isCase = myInfert$case == 1
    
    # Estimate a logistic regression model
    # With no normalization or regularization, weights should be in
    # same ball park as rxLogit
    myFormula <- isCase ~ age + parity + spontaneous + induced
    logitModel <- rxLogisticRegression(formula = myFormula,
                      data = myInfert, l1 = 0, l2 = 0, normalize = "no")
    sumLogitMx <- summary(logitModel, digits = 4)  
    coefMx <- signif(coef(sumLogitMx), digits = 4)
    checkEquals(-2.852, coefMx["(Bias)"], tolerance = 1e-03, checkName = FALSE)
    checkEquals(0.05317, coefMx["age"], tolerance = 1e-04, checkName = FALSE)
    checkEquals(-0.7089, coefMx["parity"], tolerance = 1e-03, checkName = FALSE)
    checkEquals(1.925, coefMx["spontaneous"], tolerance = 1e-03, checkName = FALSE)
    checkEquals(1.19, coefMx["induced"], tolerance = 1e-03, checkName = FALSE)
    
    logitModelL2 <- rxLogisticRegression(formula = myFormula,
                      data = myInfert, l1 = 0, l2 = 1)
    sumLogitMxL2 <- summary(logitModelL2)   
    coefL2 <- coef(sumLogitMxL2)
    checkTrue(coefL2["age"] > 0) # L2 doesn't remove weights
    
    logitModelL1 <- rxLogisticRegression(formula = myFormula,
                      data = myInfert, l1 = 1.2, l2 = 0)
    sumLogitMxL1 <- summary(logitModelL1) 
    coefL1 <- coef(logitModelL1)
    checkTrue(is.na(coefL1["age"])) # The small weight for age becomes 0 with L1   
}

"test.rxLogisticRegression.args.l1l2" <- function()
{
    if (rxIsExpressEdition())
        DEACTIVATED("Not running multi-thread tests for express edition.")

    myInfert <- infert
    myInfert$isCase = myInfert$case == 1
    
    # Estimate a logistic regression model
    # With no normalization or regularization, weights should be in
    # same ball park as rxLogit
    myFormula <- isCase ~ age + parity + spontaneous + induced
    logitModel <- rxLogisticRegression(formula = myFormula,
                      data = myInfert, l1 = 0, l2 = 0, normalize = "no", trainThreads = 8)#, parallel = 8
    sumLogitMx <- summary(logitModel)
    
    logitModelRx <- rxLogit(formula = myFormula,
                      data = myInfert) 
                
    sumLogitRx <- summary(logitModelRx)   
    coefMx <- coef(logitModel)
    coefRx <- coef(logitModelRx)
    checkEquals(coefRx["(Intercept)"], coefMx["(Bias)"], tolerance = 1E-04, checkNames = FALSE)
    checkEquals(coefRx["age"], coefMx["age"], tolerance = 1E-04)
    checkEquals(coefRx["parity"], coefMx["parity"], tolerance = 1E-04)
    checkEquals(coefRx["spontaneous"], coefMx["spontaneous"], tolerance = 1E-04)
    checkEquals(coefRx["induced"], coefMx["induced"], tolerance = 1E-04)
    
    logitModelL2 <- rxLogisticRegression(formula = myFormula,
                      data = myInfert, l1 = 0, l2 = 1)
    sumLogitMxL2 <- summary(logitModelL2)   
    coefL2 <- coef(sumLogitMxL2)
    checkTrue(coefL2["age"] > 0) # L2 doesn't remove weights
    
    logitModelL1 <- rxLogisticRegression(formula = myFormula,
                      data = myInfert, l1 = 1.2,l2 = 0)
    sumLogitMxL1 <- summary(logitModelL1) 
    coefL1 <- coef(logitModelL1)
    checkTrue(is.na(coefL1["age"]))  # The small weight for age becomes 0 with L1   
}

"test.rxLogisticRegression.args.env" <- function()
{
    myFun <- function(memSize = 10)
    {
        myIris <- iris
        myIris$isVersi <- iris$Species == "versicolor"    
        out <- rxLogisticRegression(isVersi~Petal.Length + Petal.Width, 
            memorySize = memSize, data = myIris)
        return(out)
    }
    x <- myFun()   
}

"test.rxLogisticRegression.formula.squiggle" <- function()
{

    fit1 <- rxLogisticRegression(Species ~ ., iris, type = "multiClass")
    checkEquals(fit1$params$depVar,  "Species")
    checkEquals(fit1$params$indVars, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width"))
    
    fit2 <- rxLogisticRegression(Species ~ ., iris, type = "multiClass", mlTransformVars = "Sepal.Width")
    checkEquals(fit2$params$depVar,  "Species")
    checkEquals(fit2$params$indVars, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width"))    

}

"test.rxLogisticRegression.do.call" <- function()
{

    fit1 <- rxLogisticRegression(Species ~ ., iris, type = "multiClass")
    sum1 <- summary(fit1)
   
    modelArgs <- 
      list(formula = as.formula("Species ~ ."),
           data = iris,
          type = "multiClass"
    )

    # If function is not in quotes, whole function appears in summary
    fit2 <- do.call(rxLogisticRegression, modelArgs)

    predOut1 <- rxPredict(fit1, data = iris)
    predOut2 <- rxPredict(fit2, data = iris)
    checkEquals(predOut1, predOut2)
    
    predArgs <- list(fit2, data = iris)
    predOut22 <- do.call(rxPredict, predArgs)
    checkEquals(predOut22, predOut2)
    
    fit3 <- do.call("rxLogisticRegression", modelArgs)
    sum3 <- summary(fit3)
    sum3$call$data <- as.symbol("iris") # Contains contents of data frame
    checkEquals(sum1, sum3)

    sum4 <- summary(fit3, top = 3)
    checkEquals(length(coef(sum4)), 3)
    checkEquals(coef(sum3)[1:3], coef(sum4))
}
 
"test.rxLogisticRegression.args.hidden" <- function()
{
    myIris <- iris
    myIris$isVersi <- iris$Species == "versicolor"    
    out <- rxLogisticRegression(isVersi~Petal.Length + Petal.Width, 
            randomSeed = 5, data = myIris)

    CheckStringsFound(out$mamlCode, "randomSeed=5")
}

"test.rxLogisticRegression.data.NA.ExpressEdition" <- function() 
{
    utils.run.ExpressEdition(run.rxLogisticRegression.data.NA.ExpressEdition)
}

"run.rxLogisticRegression.data.NA.ExpressEdition" <- function()
{
    #am is short for "automatic or manual" (and vs means "V" or "Straight").
    #0 means a V-engine, and 1 straight engine
    #am = 0 is automatic and 1 is manual.    

    # cyl : number of cylinders
    # hp : gross horsepower
    # wt : #/1000

    # Verifies that we are really running under Express Edition.
    checkTrue(rxIsExpressEdition())

    oldCars <- mtcars
    oldCars$isManual <- as.logical(oldCars$am)
    myFormula <- isManual ~ hp + wt
    lrModel <- rxLogisticRegression(myFormula, data = oldCars, norm = "No",
        l1Weight = 0, l2Weight = 0)
    sumMx <- summary(lrModel, digits = 4)
    coefMx <- coef(sumMx)
    expectedCoefMx <- c(-8.086, 0.03625, 18.87)
    checkEquals(expectedCoefMx, signif(unname(sort(unlist(coefMx))), digits=4))

    oldCars1 <- oldCars
    oldCars1$isManual[25] <- NA
    lrModel1 <- rxLogisticRegression(myFormula, data = oldCars1,
        l1Weight = 0, l2Weight = 0, norm = "no")
    sumMx <- summary(lrModel1, digits = 4)
    coefMx <- coef(sumMx)
    expectedCoefMx <- c(-8.064, 0.03618, 18.82)
    checkEquals(expectedCoefMx, signif(unname(sort(coefMx)), digits=4))
}

"test.rxLogisticRegression.data.NA" <- function()
{
    if (rxIsExpressEdition())
        DEACTIVATED("Not running multi-thread testing for express edition.")
    
    #am is short for "automatic or manual" (and vs means "V" or "Straight").
    #0 means a V-engine, and 1 straight engine
    #am = 0 is automatic and 1 is manual.    

    # cyl : number of cylinders
    # hp : gross horsepower
    # wt : #/1000
        
    oldCars <- mtcars    
    oldCars$isManual <- as.logical(oldCars$am)
    myFormula <- isManual~ hp + wt
    rxOut0 <- rxLogit(myFormula, data = oldCars)
    coefRx <- coef(rxOut0)
    names(coefRx) <- c("(Bias)", "hp", "wt")
    lrModel <- rxLogisticRegression(myFormula, data = oldCars, norm = "No",
        l1Weight = 0, l2Weight = 0)
    sumMx <- summary(lrModel)
    coefMx <- coef(sumMx)
    checkEquals(sort(coefRx), sort(coefMx), tolerance = 1E-04)
    
    oldCars1 <- oldCars
    oldCars1$isManual[25] <- NA
    rxOut1 <- rxLogit(myFormula, data = oldCars1)
    coefRx <- coef(rxOut1)
    names(coefRx) <- c("(Bias)", "hp", "wt")
    lrModel1 <- rxLogisticRegression(myFormula, data = oldCars1,
        l1Weight = 0, l2Weight = 0, norm = "no")
    sumMx <- summary(lrModel1)
    coefMx <- coef(sumMx)
    checkEquals(sort(coefRx), sort(coefMx), tolerance = 1E-04)
}

"test.rxLogisticRegression.args.type.multiClass" <- function()
{
	# Create temporary data sources
    myIris = iris
    
    myXdf <- tempfile(pattern = "myIris", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)
    rxDataStep(myIris, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)
    
    myCsv <- tempfile(pattern = "myIris", fileext = ".csv")
    on.exit(file.remove(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, rowsPerRead = 40, 
        colInfo = list(Species = list(type="factor", levels = c("setosa", "versicolor", "virginica" ))))
    rxDataStep(myIris, myCsvDS, overwrite = TRUE)
    
    readVars <- c("Species", "Sepal.Length", "Sepal.Width",  "Petal.Width") 
    # Run examples
    res1 <- rxLogisticRegression(formula = Species~Sepal.Length + Sepal.Width + Petal.Width,
       type = "multiClass", data = myXdfDS)
    checkEquals(res1$params$readVars, readVars)
    checkEquals(res1$coefficients, coef(res1))
    #scoreData <- rxPredict(res1, data = myXdfDS, outData = NULL, extraVarsToWrite = "Species")
    #checkTrue(sum(abs(scoreData$Species - scoreData$PredictedLabel)) <= 7)
    
    checkTrue(is.vector(coef(res1)))
    
    res2 <- rxLogisticRegression(formula = Species~Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris,  type = "multi")
    checkEquals(res2$params$readVars, readVars)
    
    res3 <- rxLogisticRegression(formula = Species~Sepal.Length + Sepal.Width + Petal.Width,
        data = myCsvDS,  type = "multi")
    checkEquals(res3$params$readVars, readVars) 

}

"test.rxLogisticRegression.data.types" <- function()
{
    inFile <- file.path(rxGetOption("sampleDataDir"), "claims.txt")
    dataColInfo <- list(
        RowNum =list(type = "integer"), #ok for integer but failed for int16 and uint16
        age = list( type = "factor",
            levels = c("17-20", "21-24", "25-29", "30-34", "35-39", "40-49", "50-59", "60+")),
        car.age = list(type = "factor",
         levels = c("0-3", "4-7", "8-9", "10+")),
        type = list(type = "factor",
            levels = c("A", "B", "C", "D")))
    inData <- RxTextData(inFile, colInfo = dataColInfo)
    form <- type ~ cost + number + RowNum
    model1 <- rxLogisticRegression(form, data = inData, type = "multiClass")
	sum1 <- summary(model1)
    
	# int16
	dataColInfo$RowNum$type <- "int16"
    model2 <- rxLogisticRegression(form, data = inData, type = "multiClass")
    sum2 <- summary(model2)
	checkEquals(sum1, sum2)	
	    
	# uint16
	dataColInfo$RowNum$type <- "uint16"
    model3 <- rxLogisticRegression(form, data = inData, type = "multiClass")
    sum3 <- summary(model3)
	checkEquals(sum1, sum3)		    
	
	# int64
	dataColInfo$RowNum$type <- "int64"
    model4 <- rxLogisticRegression(form, data = inData, type = "multiClass")
    sum4 <- summary(model4)
	checkEquals(sum1, sum4)			    
	
}


"test.rxLogisticRegression.binary.xdf.multiblock" <- function()
{

	inData <- file.path(rxGetOption("sampleDataDir"), "claims4blocks.xdf")
	form <- (car.age == "8-9") ~ age + type + cost + number
	
	###################################
	## MicrosoftML
	
    ## train
    outModelRML <- rxLogisticRegression(form, data = inData)
	
	## rxPredict
	#outFileRML <- NULL    #will write back to inData!
    #outFileRML <- tempfile(pattern = "outScore", fileext = ".txt")
    outFileRML <- tempfile(pattern = "outScore", fileext = ".xdf")
	on.exit(file.remove(outFileRML), add = TRUE)
	
    outScoreRML <- rxPredict(outModelRML, data = inData,
		outData = outFileRML, extraVarsToWrite = "RowNum")
	outDataRML <- rxDataStep(outFileRML)    #NOTE: outDataRML['Probability'] = 1 / (1 + exp(-outDataRML['Score']))
	rxGetInfo(outFileRML, getVarInfo = TRUE, getBlockSizes = TRUE)  # not multiblock even if inData is?

	###################################
	## RevoScaleR
	
	## train
    outModelRSR <- rxLogit(form, data = inData)

	## rxPredict
	#outFileRSR <- NULL
    #outFileRSR <- tempfile(pattern = "outScore", fileext = ".txt")
    outFileRSR <- tempfile(pattern = "outScore", fileext = ".xdf")
	on.exit(file.remove(outFileRSR), add = TRUE)
	
    outScoreRSR <- rxPredict(outModelRSR, data = inData,
        outData = outFileRSR, overwrite = TRUE, extraVars = "RowNum", type = "response")    #c("response", "link")
	outDataRSR <- rxDataStep(outFileRSR)
	rxGetInfo(outFileRSR, getVarInfo = TRUE, getBlockSizes = TRUE)  # not composite even if inData is?
	
	###################################
	## OSR
	
	## train
    outModelOSR <- glm(form, data = rxDataStep(inData), family = binomial(link = "logit"))
	
	## predict
    outDataOSR <- predict(outModelOSR, data = rxDataStep(inData), na.action = na.pass,
		type = "response")    #c("link", "response", "terms")
	
	###################################
    ## compare results
    ## NOTE: the behavior of predict in OSR seems changed for 3.3.2, 
    ## which always omits obs with missing values instead of predicting 'NA'!
    #cbind(outDataRML, outDataRSR, outDataOSR)   # outDataRSR matches outDataOSR but not outDataRML?
    checkTrue(TRUE)
}

"test.rxLogisticRegression.binary.xdf.composite" <- function()
{
	inData <- file.path(rxGetOption("sampleDataDir"), "claimsXdf")  #composite XDF
	#inData <- file.path(rxGetOption("sampleDataDir"), "claims4blocks.xdf")
	form <- (car.age == "8-9") ~ age + type + cost + number
	
	###################################
	## MicrosoftML
	
    ## train
    outModelRML <- rxLogisticRegression(form, data = inData)
	
	## rxPredict
	#outFileRML <- NULL    #will write back to inData!
    #outFileRML <- tempfile(pattern = "outScore", fileext = ".txt")
    outFileRML <- RxXdfData(tempfile(pattern = "outScore", fileext = "Xdf"), 
        createCompositeSet = TRUE, blocksPerCompositeFile = 4)
	on.exit(unlink(outFileRML@file, recursive = TRUE), add = TRUE)
	
    outScoreRML <- rxPredict(outModelRML, data = inData,
		outData = outFileRML, extraVarsToWrite = "RowNum")
	outDataRML <- rxDataStep(outFileRML)    #NOTE: outDataRML['Probability'] = 1 / (1 + exp(-outDataRML['Score']))
	rxGetInfo(outFileRML, getVarInfo = TRUE, getBlockSizes = TRUE)  # not multiblock even if inData is?

	###################################
	## RevoScaleR
	
	## train
    outModelRSR <- rxLogit(form, data = inData)

	## rxPredict
	#outFileRSR <- NULL
    #outFileRSR <- tempfile(pattern = "outScore", fileext = ".txt")
    outFileRSR <- RxXdfData(tempfile(pattern = "outScore", fileext = "Xdf"),
        createCompositeSet = TRUE, blocksPerCompositeFile = 4)
	on.exit(unlink(outFileRSR@file, recursive = TRUE), add = TRUE)
	
    outScoreRSR <- rxPredict(outModelRSR, data = inData,
        outData = outFileRSR, overwrite = TRUE, extraVars = "RowNum", type = "response")    #c("response", "link")
	outDataRSR <- rxDataStep(outFileRSR)
	rxGetInfo(outFileRSR, getVarInfo = TRUE, getBlockSizes = TRUE)  # not composite even if inData is?

    checkTrue(TRUE)
}

"test.rxLogisticRegression.binary.xdf.multiblock.rowSelection" <- function()
{
	inData <- file.path(rxGetOption("sampleDataDir"), "claims4blocks.xdf")
	form <- (car.age == "8-9") ~ age + type + cost + number

    ## no empty block
    outModelRML <- rxLogisticRegression(form, data = inData)

    ## the first block is empty after rowSelection
    outModelRML <- rxLogisticRegression(form, data = inData
        , rowSelection = (RowNum > 32)
        , normalize = "no"
        , trainThreads = 1,
        , maxIterations = 1
        )

    ## the middle blocks are empty after rowSelection
    outModelRML <- rxLogisticRegression(form, data = inData, rowSelection = (RowNum <= 32 | RowNum > 96)
        , normalize = "no"
        , trainThreads = 1
        , maxIterations = 1
        )

    ## the last block is empty after rowSelection
    outModelRML <- rxLogisticRegression(form, data = inData, rowSelection = (RowNum <= 96)
        , normalize = "no"
        , trainThreads = 1
        , maxIterations = 1
        )

    checkTrue(TRUE)
}

"test.rxLogisticRegression.binary.xdf.composite" <- function()
{
    claimsTxt <- file.path(rxGetOption("sampleDataDir"), "claims.txt")
    claimsTxtDs <- RxTextData(claimsTxt,  stringsAsFactors = TRUE)
    claimsDf <- rxImport(claimsTxtDs)
    
    numObs <- nrow(claimsDf)
    numXdfd <- 4L
    blocksPerCompositeFile <- 4L
    (blockSize <- as.integer(numObs / (numXdfd * blocksPerCompositeFile)))
    offset <- 2 # to make the row distribution uneven across the blocks
    createCompositeSet <- TRUE  #FALSE  #
	
    compXdf1 <- tempfile( pattern = "compXdf1", fileext = "")
    on.exit(if(file.exists(compXdf1)) unlink(compXdf1, recursive = TRUE, force = TRUE), add = TRUE)
    compXdf1.ds <- RxXdfData(compXdf1, createCompositeSet = createCompositeSet)
    rxImport(inData = claimsTxt, outFile = compXdf1.ds, overwrite = TRUE, stringsAsFactors = TRUE,
        rowsPerRead = blockSize + offset, blocksPerCompositeFile = blocksPerCompositeFile    #createCompositeSet = TRUE,
        )
    compXdf1.info <- rxGetInfo(compXdf1.ds, getVarInfo = TRUE, getBlockSizes = TRUE)

	#inData <- file.path(rxGetOption("sampleDataDir"), "claimsXdf")
    inData <- compXdf1.ds
	form <- (car.age == "8-9") ~ age + type + cost + number
	
	###################################
	## MicrosoftML
	
    ## train
    outModelRML <- rxLogisticRegression(form, data = inData, l2Weight = 0, l1Weight = 0)
	
	## rxPredict
	#outFileRML <- NULL    #will write back to inData!
    #outFileRML <- tempfile(pattern = "outScore", fileext = ".txt")
    outFileRML <- tempfile(pattern = "outScore", fileext = "")
    outFileRML.ds <- RxXdfData(outFileRML, createCompositeSet = createCompositeSet,
        blocksPerCompositeFile = 4)
	on.exit(file.remove(outFileRML), add = TRUE)
	
    outScoreRML <- rxPredict(outModelRML, data = inData,
		outData = outFileRML.ds, extraVarsToWrite = "RowNum", overwrite = TRUE)
	outDataRML <- rxDataStep(outFileRML)    #NOTE: outDataRML['Probability'] = 1 / (1 + exp(-outDataRML['Score']))
	outDataRML.info <- rxGetInfo(outFileRML, getVarInfo = TRUE, getBlockSizes = TRUE)  # not multiblock even if inData is?
    checkEquals(outDataRML.info$numRows, numObs)
    
	###################################
	## RevoScaleR
	
	## train
    outModelRSR <- rxLogit(form, data = inData)

	## rxPredict
	#outFileRSR <- NULL
    #outFileRSR <- tempfile(pattern = "outScore", fileext = ".txt")
    outFileRSR <- tempfile(pattern = "outScore", fileext = ".xdf")
	on.exit(file.remove(outFileRSR), add = TRUE)
	
    outScoreRSR <- rxPredict(outModelRSR, data = inData,
        outData = outFileRSR, overwrite = TRUE, extraVars = "RowNum", type = "response")    #c("response", "link")
	outDataRSR <- rxDataStep(outFileRSR)
	outDataRSR.info <- rxGetInfo(outFileRSR, getVarInfo = TRUE, getBlockSizes = TRUE)  # not composite even if inData is?
    checkEquals(outDataRSR.info$numRows, numObs)

	###################################
	## OSR
	
	## train
    outModelOSR <- glm(form, data = rxDataStep(inData), family = binomial(link = "logit"))
	
    ## predict
    ## NOTE: the behavior of predict in OSR seems changed for 3.3.2, 
    ## which always omits obs with missing values instead of predicting 'NA'!
    outDataOSR <- predict(outModelOSR, data = rxDataStep(inData),
		type = "response")    #c("link", "response", "terms")
    #checkEquals(length(outDataOSR), numObs)
    
	###################################
	## compare results
	#cbind(outDataRML, outDataRSR, outDataOSR)
    #checkEquals(outDataRSR[[1]], outDataOSR, checkNames = FALSE)
    #checkEquals(outDataRML$Probability, outDataOSR, checkNames = FALSE, tolerance = 1e-3)
    checkTrue(TRUE)
}

"test.rxLogisticRegression.multi.integer.label" <- function()
{
	nvar <- 2
	nobs <- 10
	y <- c(rep(0L, nobs), rep(4L, nobs)) 
	set.seed(123)
	offset <- 5
	c1 <- matrix(rnorm(nobs * nvar, mean = offset), nrow = nobs)
	c2 <- matrix(rnorm(nobs * nvar, mean = -offset), nrow = nobs)
	x <- rbind(c1, c2)
	cnames <- paste0("x", seq.int(nvar))
	colnames(x) <- cnames
	yx <- cbind.data.frame(y, x)
	yx
	## modeling and prediction
	form <- paste("y", paste(cnames, collapse = "+"), sep = "~")
	fit <- rxLogisticRegression(form, yx, type = "multiClass")  # or rxNeuralNet
	(pred <- rxPredict(fit, data = yx, extraVarsToWrite="y"))
	
	checkEquals(pred$y, as.numeric(as.character(pred$PredictedLabel)))
}

"test.rxLogisticRegression.methods.coef.ExpressEdition" <- function()
{
    utils.run.ExpressEdition(run.rxLogisticRegression.methods.coef.ExpressEdition)
}

"run.rxLogisticRegression.methods.coef.ExpressEdition" <- function()
{
    # Verifies that we are really running under Express Edition.
    checkTrue(rxIsExpressEdition())

    mortXdf <- file.path(rxGetOption("sampleDataDir"), "mortDefaultSmall")
    mortDF <- rxDataStep(mortXdf)

    # Create a logical TRUE/FALSE variable
    mortDF$default <- mortDF$default == 1

    # Divide the data into train and test data sets
    set.seed(37)
    isTest <- rnorm(nrow(mortDF)) > 0
    mortTrain <- mortDF[!isTest,]

    # Specify the formula to use 
    binaryFormula <- default ~ creditScore + houseAge + yearsEmploy +  
        ccDebt + year
 		
	logitModel1 <- rxLogisticRegression(binaryFormula, data = mortTrain)
    
    coefOut1 <- coef(logitModel1)
	checkEquals(sort(names(coefOut1)), sort(c("(Bias)", "ccDebt", "creditScore", "yearsEmploy", "houseAge", "year")),
            msg = paste(":", paste(names(coefOut1), collapse = ", ")))

    logitModel2 <- rxLogisticRegression(formula = binaryFormula, data = mortTrain, l2Weight = 100, 
         l1Weight = 100)
    coefOut2 <- coef(logitModel2)  
    checkEquals(names(coefOut2), "(Bias)")
}

"test.rxLogisticRegression.methods.coef" <- function()
{
    if (rxIsExpressEdition())
        DEACTIVATED("Not running multi-thread testing for express edition.")

    mortXdf <- file.path(rxGetOption("sampleDataDir"), "mortDefaultSmall")
    mortDF <- rxDataStep(mortXdf)

    # Create a logical TRUE/FALSE variable
    mortDF$default <- mortDF$default == 1

    # Divide the data into train and test data sets
    set.seed(37)
    isTest <- rnorm(nrow(mortDF)) > 0
    mortTrain <- mortDF[!isTest,]

    # Specify the formula to use 
    binaryFormula <- default ~ creditScore + houseAge + yearsEmploy + ccDebt + year
    #NOTE (lgong) 2016-12-20:  Bug 27392 rxLogisticRegression gives different results for different number of trainThreads
    logitModel1 <- rxLogisticRegression(binaryFormula, data = mortTrain, trainThreads = 1) 
    
    coefOut1 <- coef(logitModel1)
    coefOut1 <- coefOut1[abs(coefOut1) > 1e-4]  #get rid of tiny (unstable) coefficients
    checkEquals(sort(names(coefOut1)), sort(c("(Bias)", "ccDebt", "creditScore", "yearsEmploy", "houseAge")),
        msg = paste(":", paste(names(coefOut1), collapse = ", ")))

    logitModel2 <- rxLogisticRegression(formula = binaryFormula, data = mortTrain, l2Weight = 100, 
         l1Weight = 100)
    coefOut2 <- coef(logitModel2)
    checkEquals(names(coefOut2), "(Bias)")
}

"utils.run.ExpressEdition" <- function(testFunc) {
    if (.Platform$OS!="windows")
        DEACTIVATED("Only windows platfrom has express edition.")
    # Whether the test runner was already in Express Edition
    wasExpressEdition = rxIsExpressEdition()
    # Set to run this test under Express Edition
    if (!wasExpressEdition)
        rxOptions(isExpressEdition = TRUE)
    # Verifies that we are really running under Express Edition.
    checkTrue(rxIsExpressEdition())
    tryCatch(do.call(testFunc, list()), finally = 
             {
                 # Restore original test runner setting regardless of whether
                 # test has passed or failed.
                 # Would be better to get rid of this if RUnit support
                 # test initializers and finalizers.
                 rxOptions(isExpressEdition = wasExpressEdition)
                 # Verify that the restoration of Express Edition setting 
                 # is successful. 
                 checkEquals(rxIsExpressEdition(), wasExpressEdition)
             })
}
"test.rxLogisticRegression.MAML.code.check.optional.params" <- function() 
{
    myIris = iris
    myIris$IsVersi <- iris$Species == "versicolor"

    defaultMamlCode <- "lab={IsVersi} feat={Sepal.Length} feat={Sepal.Width} feat={Petal.Length} feat={Petal.Width}   normalize=Auto trainer=LogisticRegression{ l2Weight=1 l1Weight=1 optTol=0.0000001 memorySize=20 maxIterations=2147483647 showTrainingStats=- sgdInitializationTolerance=0 quiet=- initWtsDiameter=0 denseOptimizer=-} calibrator=sigmoid postTransformCache=Disk"
    allParamsChanged <- "lab={IsVersi} feat={Sepal.Length} feat={Sepal.Width} feat={Petal.Length} feat={Petal.Width}   normalize=warn trainer=LogisticRegression{ l2Weight=2 l1Weight=2 optTol=0.000001 memorySize=21 maxIterations=9999999 showTrainingStats=+ sgdInitializationTolerance=1 quiet=- initWtsDiameter=1 numThreads=1 denseOptimizer=+} calibrator=sigmoid postTransformCache=Disk"
          
    irisForm <- IsVersi ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width
    
    resDefault <- rxLogisticRegression(formula = irisForm,
        data = myIris)
    resAllParamsChanged <- rxLogisticRegression(formula = irisForm, data = myIris, l2Weight = 2, l1Weight = 2, optTol = 1E-06, memorySize = 21, 
        initWtsScale = 1, maxIterations = 9999999, showTrainingStats = TRUE, sgdInitTol = 1, trainThreads = 1, denseOptimizer = TRUE, normalize = "warn")

    checkEquals(resDefault$mamlCode, defaultMamlCode)
    checkEquals(resAllParamsChanged$mamlCode,allParamsChanged)
}

"test.rxLogisticRegression.ensemble.binary" <- function()
{
    #library(MicrosoftML)
    #rm(list=ls())

    myDf <- within(iris, {Setosa <- (Species == "setosa")})
    myFormula <- Setosa ~ Sepal.Length + Sepal.Width + Petal.Length + Sepal.Width + Species

    randomSeed <- 111
    replace <- TRUE
    sampRate <- 0.75
    combineMethod <- NULL

    # data frame
    modelCount <- 1
    resDf <- rxLogisticRegression(formula = myFormula, data = myDf
        , ensemble = ensembleControl(modelCount = modelCount, replace = replace)
        )
    checkTrue(!inherits(resDf, "Ensemble"))
    #print(structure(resDf$mamlCode, class = "maml"))

    # xdf file
    myXdf <- tempfile(pattern = "myXdf", fileext = ".xdf")
    on.exit(unlink(myXdf), add = TRUE)
    rxDataStep(myDf, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)

    modelCount <- 2
    resXdf <- rxLogisticRegression(formula = myFormula, data = myXdfDS
        , ensemble = ensembleControl(modelCount = modelCount, sampRate = sampRate)
        )
    checkTrue(inherits(resXdf, "Ensemble"))
    #cat(resXdf$mamlCode, "\n")

    # csv
    myCsv <- tempfile(pattern = "myCsv", fileext = ".csv")
    on.exit(unlink(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, colInfo = rxCreateColInfo(iris))
    rxDataStep(myDf, myCsvDS, overwrite = TRUE)

    modelCount <- 3
    resCsv <- rxLogisticRegression(formula = myFormula, data = myCsvDS  #, randomSeed = NULL
        , ensemble = ensembleControl(modelCount = modelCount, combineMethod = combineMethod, randomSeed = randomSeed)
        )
    checkTrue(inherits(resCsv, "Ensemble"))
    #cat(resCsv$mamlCode, "\n")
}

"test.rxLogisticRegression.ensemble.multiClass" <- function()
{
    #library(MicrosoftML)
    #rm(list=ls())

    myDf <- within(iris, {Setosa <- (Species == "setosa")})
    myFormula <- Species ~ Sepal.Length + Sepal.Width + Petal.Length + Sepal.Width + Setosa

    randomSeed <- 111
    replace <- TRUE
    sampRate <- 0.75
    combineMethod <- NULL
    type <- "multiClass"

    # data frame
    modelCount <- 1
    resDf <- rxLogisticRegression(formula = myFormula, data = myDf, type = type
        , ensemble = ensembleControl(modelCount = modelCount, replace = replace)
        )
    checkTrue(!inherits(resDf, "Ensemble"))
    #print(structure(resDf$mamlCode, class = "maml"))

    # xdf file
    myXdf <- tempfile(pattern = "myXdf", fileext = ".xdf")
    on.exit(unlink(myXdf), add = TRUE)
    rxDataStep(myDf, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)

    modelCount <- 2
    resXdf <- rxLogisticRegression(formula = myFormula, data = myXdfDS, type = type
        , ensemble = ensembleControl(modelCount = modelCount, sampRate = sampRate, randomSeed = randomSeed)
        )
    checkTrue(inherits(resXdf, "Ensemble"))
    #cat(resXdf$mamlCode, "\n")

    # csv
    myCsv <- tempfile(pattern = "myCsv", fileext = ".csv")
    on.exit(unlink(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, colInfo = rxCreateColInfo(iris))
    rxDataStep(myDf, myCsvDS, overwrite = TRUE)

    modelCount <- 3
    resCsv <- rxLogisticRegression(formula = myFormula, data = myCsvDS, type = type
        , ensemble = ensembleControl(modelCount = modelCount, combineMethod = combineMethod, randomSeed = randomSeed)
        )
    checkTrue(inherits(resCsv, "Ensemble"))
    #cat(resCsv$mamlCode, "\n")
}

# Helper function for unit tests.
"CheckStringsFound" <- function(outputList, stringsToFind) {
    outputConcatenated <- paste(outputList, collapse = "")
    for (string in stringsToFind) {
        checkTrue(length(grep(string, outputConcatenated)) > 0)
    }
}
