# Copyright (c) Microsoft Corporation.  All rights reserved.

# Tests for basic functionality of rxOneClassSvm
".rxSqliteConnectionStringFromFile" <- function()
{
    if (exists("rxSqliteConnectionStringFromFile", globalenv()))
        get("rxSqliteConnectionStringFromFile", globalenv())
    else
        RevoScaleR:::rxSqliteConnectionStringFromFile
}

"test.rxOneClassSvm.basic" <- function()
{
    # Create temporary data sources
    myIris = iris
    myIris$isIris <- 1
    
    myXdf <- tempfile(pattern = "myIris", fileext = ".xdf")
    on.exit(file.remove(myXdf), add = TRUE)
    rxDataStep(myIris, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)
    
    myCsv <- tempfile(pattern = "myIris", fileext = ".csv")
    on.exit(file.remove(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv)
    rxDataStep(myIris, myCsvDS, overwrite = TRUE)
    
    readVars <- c("isIris", "Sepal.Length", "Sepal.Width",  "Petal.Length", "Petal.Width") 
    # Run examples
    modelObject <- rxOneClassSvm(formula = isIris~Sepal.Length + Sepal.Width + Petal.Width,
        data = myXdfDS)
    #checkTrue(is.list(coef(modelObject)))  #TODO: what to expect
    
    irisExtra <- data.frame(
        Sepal.Length = c(2.5, 2.6),
        Sepal.Width = c(.75, .9),
        Petal.Length = c(2.5, 2.5),
        Petal.Width = c(.8, .7),
        isIris = 0)

    
    irisScore <- iris[, c("Sepal.Length", "Sepal.Width",  "Petal.Length", "Petal.Width")]
    irisScore$isIris <- 1
    irisScore <- rbind(irisScore, irisExtra)
       
    scoreDF <- rxPredict( modelObject, 
         data = irisScore, extraVarsToWrite = c("isIris", "Score"), overwrite = TRUE
         )    
   
    #rxLinePlot(isIris~Score, data = scoreDF, type = "p")
        
}

"test.rxOneClassSvm.args.formula" <- function()
{
    myIris = iris
    myIris$Species <- NULL
    
    irisExtra <- data.frame(
        Sepal.Length = c(2.5, 2.6),
        Sepal.Width = c(.75, .9),
        Petal.Length = c(2.5, 2.5),
        Petal.Width = c(.8, .7),
        isIris = 0)
 
    irisScore <- iris[, c("Sepal.Length", "Sepal.Width",  "Petal.Length", "Petal.Width")]
    irisScore$isIris <- 1
    irisScore <- rbind(irisScore, irisExtra)
        
    myFormula1 <- ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width
    myFormula2 <- ~ .
    
    res1 <- rxOneClassSvm(formula = myFormula1, data = myIris)   
    res2 <- rxOneClassSvm(formula = myFormula2, data = myIris)
    
    score1 <- rxPredict(res1, data = irisScore)
    score2 <- rxPredict(res2, data = irisScore)    
    checkEquals(score1, score2)
    
    score3 <- rxPredict(res1, data = irisScore, extraVarsToWrite = "isIris")
    score4 <- rxPredict(res2, data = irisScore, extraVarsToWrite = "isIris") 
    checkEquals(score3, score4)   
    checkTrue("isIris" %in% names(score3)) 
    
    score5 <- rxPredict(res1, data = irisScore, writeModelVars = TRUE)
    score6 <- rxPredict(res2, data = irisScore, writeModelVars = TRUE) 
    checkEquals(score5, score6)   
    checkTrue("Sepal.Length" %in% names(score5)) 
    
    score7 <- rxPredict(res1, data = irisScore, writeModelVars = TRUE, extraVarsToWrite = "isIris")
    score8 <- rxPredict(res2, data = irisScore, writeModelVars = TRUE, extraVarsToWrite = "isIris") 
    checkEquals(score7, score8)   
    checkTrue("Sepal.Length" %in% names(score7))
    checkTrue("isIris" %in% names(score7))              
 }   

"test.rxOneClassSvm.args.normalizer.env" <- function()
{
    myFun <- function(myNorm = "yes")
    {    
        myIris = iris
        myIris$isIris <- 1

        # Quick smoke test to make sure it runs   
        modelObject <- rxOneClassSvm(formula = isIris~Sepal.Length + Sepal.Width + Petal.Width,
            data = myIris, normalize = myNorm) 
        return(modelObject)
    }
    out <- myFun()
}


"test.rxOneClassSvm.args.normalizer" <- function()
{
    myIris = iris
    myIris$isIris <- 1

    # Quick smoke test to make sure legal values run
    modelObject1 <- rxOneClassSvm(formula = isIris~Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris)
    
    modelObject2 <- rxOneClassSvm(formula = isIris~Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris, normalize = "yes") 
	
    modelObject3 <- rxOneClassSvm(formula = isIris~Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris, normalize = "no") 
	
    modelObject4 <- rxOneClassSvm(formula = isIris~Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris, normalize = "auto") 
    
    # Error handling: illegal value for normalize
    errorOut <- try(  
        modelObject5 <- rxOneClassSvm(formula = isIris~Sepal.Length + Sepal.Width + Petal.Width,
            data = myIris, normalize = 5), silent = TRUE 
        )
     checkTrue(inherits(errorOut, "try-error"))    
     checkTrue(grepl("'normalize' must be one of", as.character(errorOut)))
}

"test.rxOneClassSvm.args.kernel" <- function()
{
    set.seed(7)
    numRows <- 200
    normalData <- data.frame(
        day = 1:numRows
    )
    normalData$pageViews = runif(numRows, min = 10, max = 1000) + .5*normalData$day

    testData <- data.frame(
        day = 1:numRows
    )
    # The test data has outliers above 1000
    testData$pageViews = runif(numRows, min = 10, max = 1400) + .5*testData$day

    model1 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       kernel = rbfKernel())
    score1DF <- rxPredict(model1, data = testData, writeModelVars = TRUE)
    score1DF$groups = score1DF$Score > 0
    checkTrue(sum(score1DF$groups) > 0)
    
    model1a <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       kernel = rbfKernel(gamma = .7))
    score1aDF <- rxPredict(model1a, data = testData, writeModelVars = TRUE)
    score1aDF$groups = score1aDF$Score > 0
    checkTrue(sum(score1aDF$groups) > 0)    

    model2 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       kernel = linearKernel())
    score2DF <- rxPredict(model2, data = testData, writeModelVars = TRUE)
    score2DF$groups = score2DF$Score > 0
    checkTrue(sum(score2DF$groups) > 0)

    model3 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       kernel = polynomialKernel(a = .2))
    score3DF <- rxPredict(model3, data = testData, writeModelVars = TRUE)
    score3DF$groups = score3DF$Score > 0
    checkTrue(sum(score3DF$groups) > 0)

    model4 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       kernel = sigmoidKernel())
    score4DF <- rxPredict(model4, data = testData, writeModelVars = TRUE)
    score4DF$groups = score4DF$Score > 0
    checkTrue(sum(score4DF$groups) > 0)
}

"test.rxOneClassSvm.args.general" <- function()
{  
    dataFile <- file.path(rxGetOption("sampleDataDir"), "claims.txt")

    dataColInfo <- list(
        RowNum =list(type = "integer"),
        age = list( type = "factor",
            levels = c("17-20", "21-24", "25-29", "30-34", "35-39", "40-49", "50-59", "60+")),
        car.age = list(type = "factor",
            levels = c("0-3", "4-7", "8-9", "10+")),
        type = list(type = "factor",
            levels = c("A", "B", "C", "D")),
		cost = list(type = "float32"),
		number = list(type = "float32"))
	
	dataSource <- RxTextData(dataFile, colInfo = dataColInfo)
    
	cacheSize = 100
    kernel = rbfKernel()
    epsilon = 0.001
    nu = 0.1
    shrink = TRUE
	
	whichAge <- "60+"	
	form <- agen ~ car.age + type + cost + number
	transforms <- bquote(list(agen = !(age == .(whichAge))))
	rowSelection <- bquote(!(age == .(whichAge)))
	
    rxSetComputeContext("local")
    model <- rxOneClassSvm(form, data = dataSource,
        cacheSize = cacheSize, kernel = kernel, epsilon = epsilon, nu = nu, shrink = shrink,
        transforms = transforms, rowSelection = rowSelection
		)
	
    summ <- summary(model)
    score <- rxPredict(model, data = dataSource, extraVarsToWrite = "age")
    dataInfo <- rxGetInfo(dataSource, getVarInfo = TRUE, getBlockSizes = TRUE, computeInfo = TRUE)
    checkEquals(dataInfo$numRows, nrow(score))
}

"test.rxOneClassSvm.args.cacheSize" <- function()
{
    set.seed(14)
    numRows <- 2000
    normalData <- data.frame(
        day = 1:numRows
    )
    normalData$pageViews = runif(numRows, min = 10, max = 1000) + .5*normalData$day


    dataSizeMB <- as.numeric(object.size(normalData)/1048576)
    
    #FIXME: What is reasonable range?
    tm1 <- system.time(
    model1 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData, 
        cacheSize = -1)
    )
    tm2 <- system.time(
    model1 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       cacheSize = 10)
    )
 
}

"test.rxOneClassSvm.args.epsilon" <- function()
{
    set.seed(25)
    numRows <- 2500
    normalData <- data.frame(
        day = 1:numRows
    )
    normalData$pageViews = runif(numRows, min = 10, max = 1000) + .5*normalData$day

    out1 <- capture.output(
    model1 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData, 
        epsilon = .00000001, reportProgress = 0)
    )
    CheckStringsFound(out1, glob2rx("*eps=1e-*08*"))
    
    out2 <- capture.output(
    model2 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       epsilon = .1, reportProgress = 0)
    )
    CheckStringsFound(out2, "eps=0.1")
   
    res3 <- try(
    model3 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       epsilon = -1, reportProgress = 0), silent = TRUE
    )   

    checkTrue(inherits(res3, "try-error"))    
    checkTrue(grepl("'epsilon' must be greater than or", as.character(res3))) 
}

"test.rxOneClassSvm.args.nu" <- function()
{
    set.seed(3)
    numRows <- 2500
    normalData <- data.frame(
        day = 1:numRows
    )
    normalData$pageViews = runif(numRows, min = 10, max = 1000) + .5*normalData$day

    out1 <- capture.output(
    model1 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData, 
        nu = .001, reportProgress = 0)
    )
    CheckStringsFound(out1, "nu=0.001")
    
    out2 <- capture.output(
    model2 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       nu = .999, reportProgress = 0)
    )
    CheckStringsFound(out2, "nu=0.999")
    
    coef1 <- coef(model1)
    coef2 <- coef(model2)
    checkTrue(length(coef1) < length(coef2))
   
    res3 <- try(
    model3 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       nu = 2, reportProgress = 0), silent = TRUE
    )   

    checkTrue(inherits(res3, "try-error"))    
    checkTrue(grepl("'nu' must be less than or", as.character(res3)))     
}

"test.rxOneClassSvm.args.shrink" <- function()
{
    set.seed(5)
    numRows <- 2500
    normalData <- data.frame(
        day = 1:numRows
    )
    normalData$pageViews = runif(numRows, min = 10, max = 1000) + .5*normalData$day

    # TRUE is sometimes faster
    out1 <- capture.output(
    model1 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData, 
        shrink = TRUE, reportProgress = 0)
    )
    CheckStringsFound(out1, "shrinking=1")
    
    out2 <- capture.output(
    model2 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       shrink = FALSE, reportProgress = 0)
    )
    CheckStringsFound(out2, "shrinking=0")
   
    res3 <- try(
    model3 <- rxOneClassSvm(formula = ~pageViews + day, data = normalData,
       shrink = 2, reportProgress = 0), silent = TRUE
    )   

    checkTrue(inherits(res3, "try-error"))    
    checkTrue(grepl("'shrink' must be a logical", as.character(res3)))       
}

"test.rxOneClassSvm.MAML.codes.check.optional.params" <- function() 
{
    # Create temporary data sources
    myIris = iris
   
    defaultMamlCode <- "feat={Sepal.Length} feat={Sepal.Width} feat={Petal.Width}   normalize=Auto trainer=OneClassSVM{ kernel=RbfKernel cacheSize=100 epsilon=0.001 nu=0.1 shrink=+} calibrator=sigmoid"
    allParamsChanged <- "feat={Sepal.Length} feat={Sepal.Width} feat={Petal.Width}   normalize=warn trainer=OneClassSVM{ kernel=RbfKernel cacheSize=99 epsilon=0.002 nu=0.2 shrink=-} calibrator=sigmoid"
      
    # Run examples
    defaultModelObject <- rxOneClassSvm(formula = ~ Sepal.Length + Sepal.Width + Petal.Width,
        data = myIris)
    allParamsChangedModelObject <- rxOneClassSvm(formula = ~ Sepal.Length + Sepal.Width + Petal.Width, data = myIris, cacheSize = 99, 
        epsilon = 0.002, shrink = FALSE, normalize = "warn", verbose = 1, blocksPerRead = 100, nu = 0.2)

    checkEquals(defaultModelObject$mamlCode, defaultMamlCode)
    checkEquals(allParamsChangedModelObject$mamlCode,allParamsChanged)
}

"test.rxOneClassSvm.ensemble.anomaly" <- function()
{
    #library(MicrosoftML)
    #rm(list=ls())

    myDf <- within(iris, {Setosa <- (Species == "setosa")})
    myFormula <- ~ Sepal.Length + Sepal.Width + Petal.Length + Sepal.Width + Species

    randomSeed <- 111
    replace <- TRUE
    sampRate <- 0.75
    combineMethod <- NULL

    # data frame
    modelCount <- 1
    resDf <- rxOneClassSvm(formula = myFormula, data = myDf
        , ensemble = ensembleControl(modelCount = modelCount, replace = replace)
        )
    checkTrue(!inherits(resDf, "Ensemble"))
    #print(structure(resDf$mamlCode, class = "maml"))

    # xdf file
    myXdf <- tempfile(pattern = "myXdf", fileext = ".xdf")
    on.exit(unlink(myXdf), add = TRUE)
    rxDataStep(myDf, myXdf, rowsPerRead = 50, overwrite = TRUE)
    myXdfDS <- RxXdfData(file = myXdf)

    modelCount <- 2
    resXdf <- rxOneClassSvm(formula = myFormula, data = myXdfDS
        , ensemble = ensembleControl(modelCount = modelCount, sampRate = sampRate)
        )
    checkTrue(inherits(resXdf, "Ensemble"))
    #cat(resXdf$mamlCode, "\n")

    # csv
    myCsv <- tempfile(pattern = "myCsv", fileext = ".csv")
    on.exit(unlink(myCsv), add = TRUE)
    myCsvDS <- RxTextData(file = myCsv, colInfo = rxCreateColInfo(iris))
    rxDataStep(myDf, myCsvDS, overwrite = TRUE)

    modelCount <- 3
    resCsv <- rxOneClassSvm(formula = myFormula, data = myCsvDS  #, randomSeed = NULL
        , ensemble = ensembleControl(modelCount = modelCount, combineMethod = combineMethod, randomSeed = randomSeed)
        )
    checkTrue(inherits(resCsv, "Ensemble"))
    #cat(resCsv$mamlCode, "\n")
}

# Helper function for unit tests.
"CheckStringsFound" <- function(outputList, stringsToFind) {
    outputConcatenated <- paste(outputList, collapse = "")
    for (string in stringsToFind) {
        checkTrue(length(grep(string, outputConcatenated)) > 0)
    }
}
